"""
Collection of the messages directed to the user.
How it works is quite magical - the lambdas prevent the f-strings from being prematuraly evaluated, and are only evaluated once returned.
Fun fact: f-strings are firmly rooted in the AST.
"""

import webbrowser
from collections import defaultdict, namedtuple
from collections.abc import Callable
from enum import Enum
from pathlib import Path
from shutil import copy
from statistics import geometric_mean, median, stdev

from beartype import beartype
from jinja2 import Environment, FileSystemLoader, select_autoescape

import use
from use import __version__, config, home
from use.hash_alphabet import hexdigest_as_JACK
from use.pydantics import PyPI_Release, Version

env = Environment(
    loader=FileSystemLoader(Path(__file__).parent / "templates"), autoescape=select_autoescape()
)


def std(times):
    return stdev(times) if len(times) > 1 else 0


def _web_tinny_profiler(timings):
    copy(Path(__file__).absolute().parent / r"templates/profiling.css", home / "profiling.css")
    DAT = namedtuple("DECORATOR", "qualname name min geom_mean median stdev len total")
    timings_ = [
        DAT(
            getattr(
                func, "__qualname__", getattr(func, "__module__", "") + getattr(func, "__name__", str(func))
            ),
            func.__name__,
            min(times),
            geometric_mean(times),
            median(times),
            std(times),
            len(times),
            sum(times),
        )
        for func, times in timings.items()
    ]
    with open(home / "profiling.html", "w", encoding="utf-8") as file:
        args = {
            "timings": timings_,
        }
        file.write(env.get_template("profiling.html").render(**args))
    webbrowser.open(f"file://{home}/profiling.html")


def _web_aspectized(decorators, functions):
    copy(Path(__file__).absolute().parent / r"templates/aspects.css", home / "aspects.css")
    DAT = namedtuple("DECORATOR", "name func")
    redecorated = []
    for ID, funcs in decorators.items():
        name = functions[ID][-1].__name__
        redecorated.append(DAT(name, funcs))

    with open(home / "aspects.html", "w", encoding="utf-8") as file:
        args = {
            "decorators": redecorated,
            "functions": functions,
        }
        file.write(env.get_template("aspects.html").render(**args))
    webbrowser.open(f"file://{home}/aspects.html")


@beartype
def _web_aspectized_dry_run(
    *, decorator: Callable, hits: list, check: Callable, pattern: str, module_name: str
):
    copy(Path(__file__).absolute().parent / r"templates/aspects.css", home / "aspects.css")
    with open(home / "aspects_dry_run.html", "w", encoding="utf-8") as file:
        args = {
            "decorator": decorator,
            "hits": hits,
            "check": check,
            "module_name": module_name,
        }
        file.write(env.get_template("aspects_dry_run.html").render(**args))
    webbrowser.open(f"file://{home}/aspects_dry_run.html")


def _web_pebkac_no_version_no_hash(*, name, package_name, version, no_browser: bool):
    if not no_browser:
        webbrowser.open(f"https://snyk.io/advisor/python/{package_name}")
    return f"""Please specify version and hash for auto-installation of {package_name!r}.
{"" if no_browser else "A webbrowser should open to the Snyk Advisor to check whether the package is vulnerable or malicious."}
If you want to auto-install the latest version, try the following line to select all viable hashes:
use("{name}", version="{version!s}", modes=use.auto_install)"""


@beartype
def _web_pebkac_no_hash(
    *,
    name: str,
    package_name: str,
    version: Version,
    releases: list[PyPI_Release],
):
    copy(Path(__file__).absolute().parent / r"templates/stylesheet.css", home / "stylesheet.css")
    entry = namedtuple("Entry", "python platform hash_name hash_value jack_value")
    table = defaultdict(lambda: [])
    for rel in (rel for rel in releases if rel.version == version):
        for hash_name, hash_value in rel.digests.items():
            if hash_name not in (x.name for x in use.Hash):
                continue
            table[hash_name].append(
                entry(
                    rel.python_tag,
                    rel.platform_tag,
                    hash_name,
                    hash_value,
                    hexdigest_as_JACK(hash_value),
                )
            )

    with open(home / "web_exception.html", "w", encoding="utf-8") as file:
        args = {
            "name": name,
            "package_name": package_name,
            "version": version,
            "table": table,
        }
        file.write(env.get_template("hash-presentation.html").render(**args))

    # from base64 import b64encode
    # def data_uri_from_html(html_string):
    #    return f'data:text/html;base64,{b64encode(html_string.encode()).decode()}'
    webbrowser.open(f"file://{home}/web_exception.html")


class UserMessage(Enum):
    not_reloadable = (
        lambda name: f"Beware {name} also contains non-function objects, it may not be safe to reload!"
    )
    couldnt_connect_to_db = (
        lambda e: f"Could not connect to the registry database, please make sure it is accessible. ({e})"
    )
    use_version_warning = (
        lambda max_version: f"""Justuse is version {Version(__version__)}, but there is a newer version {max_version} available on PyPI.
To find out more about the changes check out https://github.com/amogorkon/justuse/wiki/What's-new
Please consider upgrading via
python -m pip install -U justuse
"""
    )
    cant_use = (
        lambda thing: f"Only pathlib.Path, yarl.URL and str are valid sources of things to import, but got {type(thing)}."
    )
    web_error = (
        lambda url, response: f"Could not load {url} from the interwebs, got a {response.status_code} error."
    )
    no_validation = (
        lambda url, hash_algo, this_hash: f"""Attempting to import from the interwebs with no validation whatsoever!
To safely reproduce:
use(use.URL('{url}'), hash_algo=use.{hash_algo}, hash_value='{this_hash}')"""
    )
    version_warning = (
        lambda package_name, target_version, this_version: f"{package_name} expected to be version {target_version}, but got {this_version} instead"
    )
    ambiguous_name_warning = (
        lambda package_name: f"Attempting to load the pkg '{package_name}', if you rather want to use the local module: use(use._ensure_path('{package_name}.py'))"
    )
    pebkac_missing_hash = (
        lambda *, name, package_name, version, recommended_hash, no_browser: f"""Failed to auto-install {package_name!r} because hashes aren't specified.
        {"" if no_browser else "A webbrowser should open with a list of available hashes for different platforms for you to pick."}"
        If you want to use the package only on this platform, this should work:
    use("{name}", version="{version!s}", hashes={recommended_hash!r}, modes=use.auto_install)"""
    )
    pebkac_unsupported = (
        lambda package_name: f"We could not find any version or release for {package_name} that could satisfy our requirements!"
    )
    pip_json_mess = (
        lambda package_name, target_version: f"Tried to auto-install {package_name} {target_version} but failed because there was a problem with the JSON from PyPI."
    )
    pebkac_no_version_no_hash = _web_pebkac_no_version_no_hash
    cant_import = (
        lambda package_name: f"No pkg installed named {package_name} and auto-installation not requested. Aborting."
    )
    cant_import_no_version = (
        lambda package_name: f"Failed to auto-install '{package_name}' because no version was specified."
    )

    no_distribution_found = (
        lambda package_name, version, last_version: f"Failed to find any distribution for {package_name} version {version} that can be run on this platform. (For your information, the most recent version of {package_name} is {last_version})"
    )

    no_recommendation = (
        lambda package_name, version: f"We could not find any release for {package_name} {version} that appears to be compatible with this platform. Check your browser for a list of hashes and select manually."
    )
    bad_version_given = (
        lambda package_name, version: f"{package_name} apparently has no version {version}, please check your spelling."
    )


class StrMessage(UserMessage):
    cant_import = (
        lambda package_name: f"No pkg installed named {package_name} and auto-installation not requested. Aborting."
    )


class TupleMessage(UserMessage):
    pass


class KwargMessage(UserMessage):
    pass


def _web_aspectizing_overview(*, decorator, check, pattern, visited, hits):
    msg = f"""
<html>
<head>
<script src="/brython.js"></script>
</head>
<body>
<ul>
    {''.join(f"<li>{h}</li>" for h in hits)}
</ul>
</body>
</html>
"""

    with open(use.home / "aspectizing_overview.html", "w") as f:
        f.write(msg)
    if not config.testing:
        webbrowser.open(use.home / "aspectizing_overview.html")
    return msg



"""
Module to hold the decorators and other utility functions used in justuse.
"""

import ast
import inspect

from collections.abc import Callable
from functools import singledispatch
from functools import update_wrapper
from itertools import takewhile
from textwrap import dedent

from use import NirvanaWarning


class _PipeTransformer(ast.NodeTransformer):
    def visit_BinOp(self, node):
        if not isinstance(node.op, (ast.LShift, ast.RShift)):
            return node
        if not isinstance(node.right, ast.Call):
            return self.visit(
                ast.Call(
                    func=node.right,
                    args=[node.left],
                    keywords=[],
                    starargs=None,
                    kwargs=None,
                    lineno=node.right.lineno,
                    col_offset=node.right.col_offset,
                )
            )
        node.right.args.insert(0 if isinstance(node.op, ast.RShift) else len(node.right.args), node.left)
        return self.visit(node.right)

def pipes(func_or_class):
    if inspect.isclass(func_or_class):
        decorator_frame = inspect.stack()[1]
        ctx = decorator_frame[0].f_locals
        first_line_number = decorator_frame[2]
    else:
        ctx = func_or_class.__globals__
        first_line_number = func_or_class.__code__.co_firstlineno
    source = inspect.getsource(func_or_class)
    tree = ast.parse(dedent(source))
    ast.increment_lineno(tree, first_line_number - 1)
    source_indent = sum(1 for _ in takewhile(str.isspace, source)) + 1
    for node in ast.walk(tree):
        if hasattr(node, "col_offset"):
            node.col_offset += source_indent
    tree.body[0].decorator_list = [
        d
        for d in tree.body[0].decorator_list
        if isinstance(d, ast.Call) and d.func.id != "pipes" or isinstance(d, ast.Name) and d.id != "pipes"
    ]
    tree = _PipeTransformer().visit(tree)
    code = compile(tree, filename=(ctx["__file__"] if "__file__" in ctx else "repl"), mode="exec")
    exec(code, ctx)
    return ctx[tree.body[0].name]

# singledispatch for methods
def methdispatch(func):
    dispatcher = singledispatch(func)

    def wrapper(*args, **kwargs):

        # so we can dispatch on None
        if len(args) == 1:
            if not kwargs:
                raise NirvanaWarning("No use trying to use Nothing.")
            args = args + (None,)
        return dispatcher.dispatch(args[1].__class__)(*args, **kwargs)

    wrapper.register = dispatcher.register
    update_wrapper(wrapper, func)
    return wrapper



"""
Delegating package installation to pip, packaging and friends.
"""

import codecs
import collections
import contextlib
import importlib.util
import inspect
import linecache
import os
import platform
import re
import sys
import tarfile
import time
import traceback
import zipfile
import zipimport
from collections.abc import Callable
from functools import lru_cache as cache
from functools import reduce
from importlib import metadata
from importlib.machinery import ModuleSpec, SourceFileLoader
from importlib.metadata import Distribution
from itertools import chain, product, zip_longest
from logging import getLogger
from numbers import Complex, Integral, Number, Rational, Real
from pathlib import Path, PureWindowsPath, WindowsPath
from shutil import rmtree
from sqlite3 import Cursor
from subprocess import CalledProcessError, run
from types import ModuleType
from typing import Any, Optional, Protocol, TypeVar, Union, get_args, get_origin, runtime_checkable
from warnings import catch_warnings, filterwarnings, warn

import furl
import requests
from beartype import beartype
from furl import furl as URL
from icontract import ensure, require
from packaging import tags
from packaging.specifiers import SpecifierSet

from use import Hash, InstallationError, Modes, UnexpectedHash, VersionWarning, config
from use.hash_alphabet import hexdigest_as_JACK, num_as_hexdigest
from use.messages import UserMessage, _web_pebkac_no_hash
from use.pydantics import PyPI_Project, PyPI_Release, PyPI_URL, RegistryEntry, Version
from use.tools import pipes

log = getLogger(__name__)


T = TypeVar("T")


class PlatformTag:
    def __init__(self, platform: str):
        self.platform = platform

    def __str__(self):
        return self.platform

    def __repr__(self):
        return f"use.PlatformTag({self.platform!r})"

    def __hash__(self):
        return hash(self.platform)

    @require(lambda self, other: isinstance(other, self.__class__))
    def __eq__(self, other):
        return self.platform == other.platform


@beartype
def _ensure_version(
    result: Union[ModuleType, Exception], *, name, version, **kwargs
) -> Union[ModuleType, Exception]:
    if not isinstance(result, ModuleType):
        return result
    result_version = _get_version(mod=result)
    if result_version != version:
        warn(UserMessage.version_warning(name, version, result_version), category=VersionWarning)
    return result


# fmt: off
@pipes
def _ensure_path(value: Union[bytes, str, furl.Path, Path]) -> Path:
    if isinstance(value, (str, bytes)):
        return Path(value).absolute()
    if isinstance(value, furl.Path):
        return (
            Path.cwd(),
            value.segments
            << map(Path)
            << tuple
            << reduce(Path.__truediv__),
        ) << reduce(Path.__truediv__)
    return value
# fmt: on


@cache
@beartype
def get_supported() -> frozenset[PlatformTag]:  # cov: exclude
    """
    Results of this function are cached. They are expensive to
    compute, thanks to some heavyweight usual players
    (*ahem* pip, package_resources, packaging.tags *cough*)
    whose modules are notoriously resource-hungry.

    Returns a set containing all platform
    supported on the current system.
    """
    get_supported = None
    with catch_warnings():
        filterwarnings(action="ignore", category=DeprecationWarning)
        with contextlib.suppress(ImportError):
            from pip._internal.resolution.legacy.resolver import get_supported
        if not get_supported:
            with contextlib.suppress(ImportError):
                from pip._internal.models.target_python import get_supported
        if not get_supported:
            with contextlib.suppress(ImportError):
                from pip._internal.utils.compatibility_tags import get_supported
        if not get_supported:
            with contextlib.suppress(ImportError):
                from pip._internal.resolution.resolvelib.factory import get_supported
    get_supported = get_supported or (lambda: [])

    items: list[PlatformTag] = [PlatformTag(platform=tag.platform) for tag in get_supported()]

    # yay backwards compatibility..
    if hasattr(tags, "platform_tags"):
        items.extend(PlatformTag(platform=str(tag)) for tag in tags.platform_tags())
    else:
        items.extend(PlatformTag(platform=str(tag)) for tag in tags._platform_tags())

    return frozenset(items)


@beartype
def _filter_by_version(releases: list[PyPI_Release], *, version: Version) -> list[PyPI_Release]:
    return list(filter(lambda r: r.version == version, releases))


class ZipFunctions:
    def __init__(self, artifact_path):
        self.archive = zipfile.ZipFile(artifact_path)

    def get(self):
        return (self.archive, [e.filename for e in self.archive.filelist])

    def read_entry(self, entry_name):
        with self.archive.open(entry_name) as f:
            bdata = f.read()
            text = bdata.decode("ISO-8859-1").splitlines() if len(bdata) < 8192 else ""
            return (Path(entry_name).stem, text)


class TarFunctions:
    def __init__(self, artifact_path):
        self.archive = tarfile.open(artifact_path)

    def get(self):
        return (self.archive, [m.name for m in self.archive.getmembers() if m.type == b"0"])

    def read_entry(self, entry_name):
        m = self.archive.getmember(entry_name)
        with self.archive.extractfile(m) as f:
            bdata = f.read()
            text = bdata.decode("UTF-8").splitlines() if len(bdata) < 8192 else ""
            return (Path(entry_name).stem, text)


@beartype
@pipes
def archive_meta(artifact_path):
    DIST_PKG_INFO_REGEX = re.compile("(dist-info|-INFO|\\.txt$|(^|/)[A-Z0-9_-]+)$")

    if ".tar" in Path(str(artifact_path)).stem:
        functions = TarFunctions(artifact_path)
    else:
        functions = ZipFunctions(artifact_path)

    archive, names = functions.get()
    meta = names << filter(DIST_PKG_INFO_REGEX.search) << map(functions.read_entry) >> dict
    meta.update(
        dict(
            (lp := l.partition(": "), (lp[0].lower().replace("-", "_"), lp[2]))[-1]
            for l in meta.get("METADATA", meta.get("PKG-INFO"))
            if ": " in l
        )
    )
    name = meta.get("name", Path(artifact_path).stem.split("-")[0])
    meta["name"] = name
    if "top_level" not in meta:
        meta["top_level"] = [""]
    (
        top_level,
        name,
    ) = (meta["top_level"][0], meta["name"])
    import_name = name if top_level == name else ".".join((top_level, name))
    meta["names"] = names
    meta["import_name"] = import_name
    for relpath in sorted(
        [n for n in names if len(n) > 4 and n[-3:] == ".py"],
        key=lambda n: (
            not n.startswith(import_name),
            not n.endswith("__init__.py"),
            len(n),
        ),
    ):
        meta["import_relpath"] = relpath
        break
    else:
        meta["import_relpath"] = f"{import_name}.py"
    archive.close()
    return meta


@beartype
def _clean_sys_modules(package_name: str) -> None:
    for k in dict(
        [
            (k, v)
            for k, v in list(sys.modules.items())
            if package_name in k.split(".")
            and (
                getattr(v, "__spec__", None) is None
                or isinstance(v, (SourceFileLoader, zipimport.zipimporter))
            )
        ]
    ):
        if k in sys.modules:
            del sys.modules[k]


@beartype
def _pebkac_no_version(
    *,
    name: str,
    func: Callable[..., Union[Exception, ModuleType]] = None,
    Message: type,
    **kwargs,
) -> Union[ModuleType, Exception]:

    if func:
        result = func()
        if isinstance(result, (Exception, ModuleType)):
            return result
        assert False, f"{func}() returned {result!r}"

    return RuntimeWarning(Message.cant_import_no_version(name))


@beartype
@pipes
def _pebkac_no_hash(
    *,
    name: str,
    version: Version,
    package_name: str,
    no_browser: bool,
    Message: type,
    hash_algo: Hash,
    **kwargs,
) -> RuntimeWarning:
    releases = _get_releases_from_pypi(package_name=package_name, version=version)
    filtered = _filter_by_platform(releases, tags=get_supported())
    ordered = _sort_releases(filtered)

    if not no_browser:
        _web_pebkac_no_hash(name=name, package_name=package_name, version=version, releases=ordered)

    if not ordered:
        return RuntimeWarning(Message.no_recommendation(package_name, version))

    recommended_hash = hexdigest_as_JACK(ordered[0].digests.get(hash_algo.name))

    # for test_suggestion_works, don't remove
    print(recommended_hash)

    return RuntimeWarning(
        Message.pebkac_missing_hash(
            name=name,
            package_name=package_name,
            version=version,
            recommended_hash=recommended_hash,
            no_browser=no_browser,
        )
    )


@beartype
@pipes
def _pebkac_no_version_no_hash(
    *,
    name: str,
    package_name: str,
    no_browser: bool,
    Message: type,
    **kwargs,
) -> Exception:
    # let's try to make an educated guess and give a useful suggestion
    proj = _get_project_from_pypi(package_name=package_name)
    if isinstance(proj, Exception):
        return proj
    releases = _get_releases(proj)

    ordered = releases >> _filter_by_platform(tags=get_supported()) >> _sort_releases
    # we tried our best, but we didn't find anything that could work

    # let's try to find *anything*
    if not ordered:
        ordered = _sort_releases(releases)
        if not ordered:
            # we tried our best..
            return RuntimeWarning(Message.pebkac_unsupported(package_name))

        if not no_browser:
            version = ordered[-1].version
            _web_pebkac_no_hash(name=name, package_name=package_name, version=version, project=proj)
            return RuntimeWarning(
                Message.pebkac_no_version_no_hash(
                    name=name, package_name=package_name, version=version, no_browser=no_browser
                )
            )

    recommended_version = ordered[0].version
    # for test_suggestion_works, don't remove (unless you want to work out the details)
    print(recommended_version)

    # we found something that could work, but it may not fit to the user's requirements
    return RuntimeWarning(
        Message.pebkac_no_version_no_hash(
            name=name,
            package_name=package_name,
            version=recommended_version,
            no_browser=no_browser,
        )
    )


@beartype
def _import_public_no_install(
    *,
    module_name: str,
    **kwargs,
) -> Union[Exception, ModuleType]:
    # builtin?
    builtin = False
    try:
        metadata.PathDistribution.from_name(module_name)
    except metadata.PackageNotFoundError:  # indeed builtin!
        builtin = True
    # TODO ehhh.. builtin needs to be handled differently?

    mod = sys.modules.get(module_name)
    imported = bool(mod)

    if not mod:
        mod = importlib.import_module(module_name)

    # # ? does this solve the caching issue?
    # if not imported:
    #     del sys.modules[module_name]
    return mod


def _parse_name(name: str) -> tuple[str, str]:
    """Parse the user-provided name into a package name for installation and a module name for import.

    The package name is whatever pip would expect.
    The module name is whatever import would expect.

    Mini-DSL: / separates the package name from the module name.
    """
    if not name:
        return None, None

    def old():
        # as fallback
        match = re.match(r"(?P<package_name>[^/.]+)/?(?P<rest>[a-zA-Z0-9._]+)?$", name)
        assert match, f"Invalid name spec: {name!r}"
        names = match.groupdict()
        package_name = names["package_name"]
        rest = names["rest"]
        if not package_name:
            package_name = rest
        if not rest:
            rest = package_name
        return package_name

    package_name = None
    module_name = ""
    if "/" in name:
        if name.count("/") > 1:
            raise ImportError(
                f"Invalid name spec: {name!r}, can't have multiple / characters as package/module separator."
            )
        package_name, _, module_name = name.partition("/")
    else:
        package_name = old()
        module_name = name
    return (package_name, module_name)


@beartype
def _check_db_for_installation(*, registry=Cursor, package_name=str, version) -> Optional[RegistryEntry]:
    query = registry.execute(
        """
        SELECT
            artifact_path, installation_path, pure_python_package
        FROM distributions
        JOIN artifacts ON artifacts.id = distributions.id
        WHERE name=? AND version=?
        ORDER BY artifacts.id DESC
        """,
        (
            package_name,
            str(version),
        ),
    ).fetchone()
    return RegistryEntry(**query) if query else None


@beartype
def _auto_install(
    *,
    package_name: str,
    module_name: str,
    func: Callable[..., Union[Exception, ModuleType]] = None,
    version: Version,
    hash_algo: Hash,
    user_provided_hashes: set[int],
    registry: Cursor,
    cleanup: bool,
    **kwargs,
) -> Union[ModuleType, BaseException]:
    """Install, if necessary, the package and import the module in any possible way."""
    if func:
        result = func()
        if isinstance(result, (Exception, ModuleType)):
            return result
        else:
            raise AssertionError(f"{func!r} returned {result!r}")

    if entry := _check_db_for_installation(registry=registry, package_name=package_name, version=version):
        # is there a point in checking the hashes at this point? probably not.
        if entry.pure_python_package:
            assert entry.artifact_path.exists()
            # let's not try to catch this - since this apparently already worked *at least once* - no fallback
            return zipimport.zipimporter(entry.artifact_path).load_module(module_name)
        # else: we have an installed package, let's try to import it
        original_cwd = Path.cwd()
        assert entry.installation_path.exists()
        os.chdir(entry.installation_path)
        # with an installed package there may be weird issues we can't be sure about so let's be safe
        try:
            return _load_venv_entry(
                module_name=module_name,
                installation_path=entry.installation_path,
            )
        except BaseException as err:
            traceback.print_exc(file=sys.stderr)
            msg = err
        finally:
            os.chdir(original_cwd)
        return ImportError(msg)

    # else: we have to download the package and install it

    releases = _get_releases_from_pypi(package_name=package_name, version=version)
    if isinstance(releases, Exception):
        return releases
    # we *did* ask the user to give us hashes of artifacts that *should* work, so let's check for those.
    # We can't be sure which one of those hashes will work on this platform, so let's try all of them.

    for H in user_provided_hashes:
        log.info(f"Attempting auto-installation of <{num_as_hexdigest(H)}>...")
        url = next(
            (URL(R.url) for R in releases if H == int(R.digests[hash_algo.name], 16)),
            None,
        )

        if not url:
            return UnexpectedHash(
                f"user provided {user_provided_hashes}, do not match any of the {len(releases)} possible artifacts"
            )

        # got an url for an artifact with a hash given by the user, let's install it
        filename = url.asdict()["path"]["segments"][-1]
        artifact_path = config.packages / filename
        _download_artifact(artifact_path=artifact_path, url=url, hash_value=H, hash_algo=hash_algo)
        try:
            log.info("Attempting to install..")
            entry = _install(
                package_name=package_name,
                artifact_path=artifact_path,
                version=version,
                force_install=True,
            )
            # packages like tensorflow-gpu only need to be installed but nothing imported, so module name is empty
            log.info("Attempting to import...")
            mod = _load_venv_entry(
                module_name=module_name,
                installation_path=entry.installation_path,
            )
            log.info("Successfully imported.")
        except BaseException as err:
            msg = err  # sic
            log.error(err)
            traceback.print_exc(file=sys.stderr)
            if entry and cleanup:
                rmtree(entry.installation_path)
                assert not entry.installation_path.exists()
            continue

        _save_package_info(
            package_name=package_name,
            version=version,
            artifact_path=entry.artifact_path,
            hash_value=int(hash_algo.value(entry.artifact_path.read_bytes()).hexdigest(), 16),
            hash_algo=hash_algo,
            installation_path=entry.installation_path,
            registry=registry,
        )
        return mod
    log.critical(
        f"Could not install {package_name!r} {version!r}. Hashes that were attempted: {[num_as_hexdigest(H) for H in user_provided_hashes]}"
    )
    return ImportError(msg)


@beartype
def _save_package_info(
    *,
    registry=Cursor,
    version: Version,
    artifact_path: Path,
    installation_path: Path,
    hash_value=int,
    hash_algo: Hash,
    package_name: str,
):
    """Update the registry to contain the pkg's metadata."""
    if not registry.execute(
        f"SELECT * FROM distributions WHERE name='{package_name}' AND version='{version}'"
    ).fetchone():
        registry.execute(
            f"""
INSERT INTO distributions (name, version, installation_path, date_of_installation, pure_python_package)
VALUES ('{package_name}', '{version}', '{installation_path}', {time.time()}, {installation_path is None})
"""
        )
        registry.execute(
            f"""
INSERT OR IGNORE INTO artifacts (distribution_id, artifact_path)
VALUES ({registry.lastrowid}, '{artifact_path}')
"""
        )
        registry.execute(
            f"""
INSERT OR IGNORE INTO hashes (artifact_id, algo, value)
VALUES ({registry.lastrowid}, '{hash_algo.name}', '{hash_value}')"""
        )
    registry.connection.commit()


@beartype
@ensure(lambda url: str(url).startswith("http"))
def _download_artifact(*, artifact_path: Path, url: URL, hash_algo: Hash, hash_value: int):
    # let's check if we downloaded it already, just in case
    if (
        artifact_path.exists()
        and int(hash_algo.value(artifact_path.read_bytes()).hexdigest(), 16) == hash_value
    ):
        log.info("Artifact already downloaded. Hashes matching.")
        return
    # this should work since we just got the url from pypi itself - if this fails, we have bigger problems
    log.info("Downloading artifact from PyPI...")
    data = requests.get(url).content
    artifact_path.write_bytes(data)
    if int(hash_algo.value(artifact_path.read_bytes()).hexdigest(), 16) != hash_value:
        # let's try once again, cosmic rays and all, believing in pure dumb luck
        log.info("Artifact downloaded but hashes don't match. Trying again...")
        data = requests.get(url).content
        artifact_path.write_bytes(data)
    if int(hash_algo.value(artifact_path.read_bytes()).hexdigest(), 16) != hash_value:
        # this means either PyPI is hacked or there is a man-in-the-middle
        raise ImportError("Hashes don't match. Aborting. Something very fishy is going on.")
    log.info("Download successful.")
    return


@beartype
def _is_pure_python_package(artifact_path: Path, meta: dict) -> bool:
    return next(
        (False for n, s in product(meta["names"], importlib.machinery.EXTENSION_SUFFIXES) if n.endswith(s)),
        ".tar" not in str(artifact_path),
    )


@beartype
def _find_module_in_venv(package_name: str, version: Version, relp: str) -> Path:
    env_dir = config.venv / package_name / str(version)
    log.debug("env_dir=%s", env_dir)
    site_dirs = [
        env_dir / f"Lib{suffix}" / "site-packages"
        if sys.platform == "win32"
        else env_dir / f"lib{suffix}" / ("python%d.%d" % sys.version_info[:2]) / "site-packages"
        for suffix in ("64", "")
    ]
    log.debug("site_dirs=%s", site_dirs)
    for p in env_dir.glob("**/*"):
        log.debug("  - %s", p.relative_to(env_dir).as_posix())

    original_sys_path = sys.path
    try:
        # Need strings for sys.path to work
        sys.path = [*map(str, site_dirs), *sys.path]
        sys.path_importer_cache.clear()
        importlib.invalidate_caches()
        # sic! importlib uses sys.path for lookup
        dist = Distribution.from_name(package_name)
        log.info("dist=%s", dist)
        log.debug("dist.files=%s", dist.files)
        path_set = dist.files
        log.debug("path_set=%s", path_set)
        for path in path_set:
            log.debug("path=%s", path)
            file = dist.locate_file(path)
            log.debug("file=%s", file)
            if not file.exists():
                continue
            real_file = Path(*file.parts[: -len(path.parts)])
            log.debug("real_file=%s", real_file)
            if real_file.exists():
                return real_file

    finally:
        sys.path = original_sys_path
    raise ImportError("No module in site_dirs")


@beartype
def _install(
    *,
    package_name: str,
    version: Version = None,
    force_install=False,
    artifact_path: Path,
) -> RegistryEntry:
    """Take care of the installation."""
    meta = archive_meta(artifact_path)
    import_parts = re.split("[\\\\/]", meta["import_relpath"])
    if "__init__.py" in import_parts:
        import_parts.remove("__init__.py")
    relp: str = meta["import_relpath"]
    venv_root = config.venv / package_name / str(version)
    site_pkgs_dir = list(venv_root.rglob("site-packages"))
    if not any(site_pkgs_dir):
        force_install = True
        module_paths = []
    else:
        installation_path = site_pkgs_dir[0]
        module_paths = venv_root.rglob(f"**/{relp}")

    python_exe = Path(sys.executable)

    if not module_paths or force_install:
        # If we get here, the venv/pip setup is required.
        # we catch errors one level higher, so we don't have to deal with them here
        env = {}
        _realenv = {
            k: v
            for k, v in chain(os.environ.items(), env.items())
            if isinstance(k, str) and isinstance(v, str)
        }

        argv = [
            python_exe,
            "-m",
            "pip",
            "--disable-pip-version-check",
            "--no-color",
            "--verbose",
            "--verbose",
            "install",
            "--pre",
            "--root",
            PureWindowsPath(venv_root).drive
            if isinstance(venv_root, (WindowsPath, PureWindowsPath))
            else "/",
            "--prefix",
            str(venv_root),
            "--progress-bar",
            "ascii",
            "--prefer-binary",
            "--exists-action",
            "i",
            "--ignore-installed",
            "--no-warn-script-location",
            "--force-reinstall",
            "--no-warn-conflicts",
            artifact_path,
        ]
        try:
            setup = dict(
                executable=python_exe,
                args=[*map(str, argv)],
                bufsize=1024,
                input="",
                capture_output=False,
                timeout=45000,
                check=True,
                close_fds=True,
                env=_realenv,
                encoding="ISO-8859-1",
                errors="ISO-8859-1",
                text=True,
                shell=False,
            )
        except CalledProcessError as err:
            log.error("::".join(err.cmd, err.output, err.stdout, err.stderr))
            raise InstallationError(err) from err
        output = run(**setup)
        log.info("Installation successful.")

    installation_path = _find_module_in_venv(package_name=package_name, version=version, relp=relp)

    return RegistryEntry(
        installation_path=installation_path,
        artifact_path=artifact_path,
        pure_python_package=False,
    )


@beartype
def _load_venv_entry(*, module_name: str, installation_path: Path) -> ModuleType:
    if not module_name:
        log.info("Module name is empty, returning empty Module.")
        return ModuleType("")
    origcwd = Path.cwd()
    original_sys_path = list(sys.path)
    # TODO we need to keep track of package-specific sys-paths
    if sys.path[0] != "":
        sys.path.insert(0, "")
    try:
        os.chdir(installation_path)
        # importlib and sys.path.. bleh
        return importlib.import_module(module_name)
    except BaseException as err:
        msg = err
        log.error(msg)
        traceback.print_exc(file=sys.stderr)
    finally:
        os.chdir(origcwd)
        sys.path = original_sys_path
    raise ImportError(msg)


@beartype
def _get_project_from_pypi(*, package_name: str) -> PyPI_Project | Exception:
    # let's check if package name is correct
    url = f"https://pypi.org/pypi/{package_name}/json"
    response = requests.get(url)
    if response.status_code == 404:
        return ImportError(UserMessage.pebkac_unsupported(package_name))
    elif response.status_code != 200:
        return RuntimeWarning(UserMessage.web_error(url, response))
    return PyPI_Project(**response.json())


@beartype
def _get_releases_from_pypi(*, package_name: str, version: Version) -> list[PyPI_Release] | Exception:
    # let's check if package name is correct
    url = f"https://pypi.org/pypi/{package_name}"
    response = requests.get(url)
    if response.status_code == 404:
        return ImportError(UserMessage.pebkac_unsupported(package_name))
    elif response.status_code != 200:
        return RuntimeWarning(UserMessage.web_error(url, response))
    # let's check if the version is a thing
    url = f"https://pypi.org/pypi/{package_name}/{version}/json"
    response = requests.get(url)
    if response.status_code == 404:
        return RuntimeWarning(UserMessage.bad_version_given(package_name, version))
    # looks good, let's get the releases for this version
    urls = response.json()["urls"]
    return [PyPI_Release(**url, version=version) for url in urls]


@beartype
def _filter_by_platform(releases: list[PyPI_Release], *, tags: frozenset[PlatformTag]) -> list[PyPI_Release]:
    def compatible(info: PyPI_Release, include_sdist=False) -> bool:
        return (
            _is_platform_compatible(info, tags, include_sdist)
            and not info.yanked
            and (include_sdist or info.justuse.ext not in ("tar", "tar.gz" "zip"))
        )

    filtered = [release for release in releases if compatible(release, include_sdist=False)]
    return filtered or [release for release in releases if compatible(release, include_sdist=True)]


@beartype
def _get_releases(project: PyPI_Project) -> list[PyPI_Release]:
    return reduce(list.__add__, project.releases.values(), [])


@beartype
@pipes
def _sort_releases(releases: list[PyPI_Release]) -> list[PyPI_Release]:
    return sorted(
        releases,
        key=(
            lambda r: (
                1 - int(r.filename.endswith(".tar.gz")),
                1 - int(r.is_sdist),
                r.version,
            )
        ),
        reverse=True,
    )


@cache
def _is_version_satisfied(specifier: str, sys_version) -> bool:
    """
    SpecifierSet("") matches anything, no need to artificially
    lock down versions at this point

    @see https://warehouse.readthedocs.io/api-reference/json.html
    @see https://packaging.pypa.io/en/latest/specifiers.html
    """
    specifiers = SpecifierSet(specifier or "")
    return not specifier or sys_version in specifiers


@beartype
@pipes
def _is_platform_compatible(
    info: PyPI_Release, platform_tags: frozenset[PlatformTag], include_sdist=False
) -> bool:

    if not include_sdist and (".tar" in info.justuse.ext or info.justuse.python_tag in ("cpsource", "sdist")):
        return False

    if "win" in (info.packagetype or "unknown") and sys.platform != "win32":
        return False

    if info.platform_tag:
        if "win32" in info.platform_tag and sys.platform != "win32":
            return False
        if "macosx" in info.platform_tag and sys.platform != "darwin":
            return False

    our_python_tag = tags.interpreter_name() + tags.interpreter_version()
    supported_tags = {
        our_python_tag,
        "py3",
        "cp3",
        f"cp{tags.interpreter_version()}",
        f"py{tags.interpreter_version()}",
    }

    if info.platform_tag:
        given_platform_tags = info.platform_tag.split(".") << map(PlatformTag) >> frozenset
    else:
        return include_sdist

    if info.is_sdist and info.requires_python:
        given_python_tag = {
            our_python_tag
            for p in info.requires_python.split(",")
            if Version(platform.python_version()) in SpecifierSet(p)
        }
    else:
        given_python_tag = set(info.python_tag.split("."))

    return any(supported_tags.intersection(given_python_tag)) and (
        (info.is_sdist and include_sdist) or any(given_platform_tags.intersection(platform_tags))
    )


@beartype
def _get_version(name: Optional[str] = None, package_name=None, /, mod=None) -> Optional[Version]:
    version: Optional[Union[Callable[...], Version, Version, str]] = None
    for lookup_name in (name, package_name):
        if not lookup_name:
            continue
        try:
            if lookup_name is not None:
                meta = metadata.distribution(lookup_name)
                return Version(meta.version)
        except metadata.PackageNotFoundError:
            continue
    if not mod:
        return None
    version = getattr(mod, "__version__", version)
    if isinstance(version, str):
        return Version(version)
    version = getattr(mod, "version", version)
    if callable(version):
        version = version()
    return Version(version)


def _build_mod(
    *,
    module_name,
    code,
    initial_globals: Optional[dict[str, Any]],
    module_path,
    package_name="",
) -> ModuleType:
    mod = ModuleType(module_name)
    log.info(f"{Path.cwd()=} {package_name=} {module_name=} {module_path=}")
    mod.__dict__.update(initial_globals or {})
    mod.__file__ = str(module_path)
    mod.__path__ = [str(module_path.parent)]
    mod.__package__ = package_name
    mod.__name__ = module_name
    loader = SourceFileLoader(module_name, str(module_path))
    mod.__loader__ = loader
    mod.__spec__ = ModuleSpec(module_name, loader)
    sys.modules[module_name] = mod
    code_text = codecs.decode(code)
    # module file "<", ">" chars are specially handled by inspect
    getattr(linecache, "cache")[module_path] = (
        len(code),  # size of source code
        None,  # last modified time; None means there is no physical file
        [
            *map(lambda ln: ln + "\x0a", code_text.splitlines())
        ],  # a list of lines, including trailing newline on each
        mod.__file__,  # file name, e.g. "<mymodule>" or the actual path to the file
    )
    # not catching this causes the most irritating bugs ever!
    try:
        codeobj = compile(code, module_path, "exec")
        exec(codeobj, mod.__dict__)
    except:  # reraise anything without handling - clean and simple.
        raise
    return mod


def _fail_or_default(exception: BaseException, default: Any):
    if default is not Modes.fastfail:
        return default
    else:
        raise exception


# ### active web dev ###
# proj = None
# from shutil import copy


# def _qwer():
#     copy(Path(__file__).absolute().parent / r"templates/stylesheet.css", config.home / "stylesheet.css")
#     package_name = "pygame"
#     name = "pygame/foo"
#     version = Version("2.1.0")
#     global proj
#     if proj is None:
#         proj = _get_data_from_pypi(package_name=package_name, version=version)
#         proj = _filter_by_version(proj, version)

#     _web_pebkac_no_hash(package_name=package_name, version=version, name=name, project=proj)


def _real_path(*, path: Path, _applied_decorators: dict, landmark) -> tuple[str, str, str, Path]:
    source_dir = Path.cwd()
    # calling from another use()d module
    # let's see where we started
    main_mod = __import__("__main__")
    # there are a number of ways to call use() from a non-use() starting point
    # let's first check if we are running in jupyter
    jupyter = "ipykernel" in sys.modules
    # we're in jupyter, we use the CWD as set in the notebook
    if not jupyter and hasattr(main_mod, "__file__"):
        # problem: user wants to use.Path("some_file_in_the_same_dir")
        # so we have to figure out where the file of the calling function is.
        # but the *calling* function could also be a decorator, living completely elsewhere
        # so we have to figure out whether we're being called by a decorator first.
        # We use use.__call__ as landmark, because that's the official entry point
        # and we can count on it being called. From there we need to check whether it was aspectized.
        # If it was, we need to skip those decorators before finally get to the user code and
        # we can actually see from where we've been called.
        frame = inspect.currentframe()
        while True:
            if frame.f_code == landmark:
                break
            else:
                frame = frame.f_back
        # a few more steps..
        for _ in _applied_decorators[landmark]:
            frame = frame.f_back
        try:
            # frame is in __call__ (or rather methdispatch), we need to step two frames back
            source_dir = Path(frame.f_back.f_back.f_code.co_filename).resolve().parent
        # we are being called from a shell like thonny, so we have to assume cwd
        except OSError:
            source_dir = Path.cwd()

    if source_dir is None:
        if main_mod.__loader__ and hasattr(main_mod.__loader__, "path"):
            source_dir = _ensure_path(main_mod.__loader__.path).parent
        else:
            source_dir = Path.cwd()
    if not source_dir.joinpath(path).exists():
        if files := [*[*source_dir.rglob(f"**/{path}")]]:
            source_dir = _ensure_path(files[0]).parent
        else:
            source_dir = Path.cwd()
    if not source_dir.exists():
        raise NotImplementedError("Can't determine a relative path from a virtual file.")
    if not path.exists():
        path = source_dir.joinpath(path).resolve()
    if not path.exists():
        raise ImportError(f"Sure '{path}' exists?")
    if not path.is_absolute():
        path = path.resolve()
    try:
        name = path.relative_to(source_dir)
    except ValueError:
        source_dir = path.parent
        os.chdir(source_dir)
        name = path.relative_to(source_dir)
    ext = name.as_posix().rpartition(".")[-1]
    name_as_path_with_ext = name.as_posix()
    name_as_path = name_as_path_with_ext[: -len(ext) - (1 if ext else 0)]
    name = name_as_path.replace("/", ".")
    name_parts = name.split(".")
    package_name = ".".join(name_parts[:-1])
    module_name = path.stem  # sic!

    return name, module_name, package_name, path


numerics = [bool, Integral, Rational, Real, Complex]


class NotComparableWarning(UserWarning):
    pass


def _modules_are_compatible(pre, post):
    for name, pre_obj in pre.__dict__.items():
        if callable(pre_obj):
            try:
                post_obj = getattr(post, name)
            except AttributeError:
                return False
            if _is_compatible(pre_obj, post_obj):
                continue
            else:
                return False
    return True


def _is_compatible(pre, post):
    sig = inspect.signature(pre)
    pre_sig = []
    # first separate args and kwargs so we can sort kwargs alphabetically
    args = []
    kwargs = []

    for k, v in sig.parameters.items():
        if v.kind is v.VAR_KEYWORD or v.POSITIONAL_OR_KEYWORD:
            kwargs.append((k, v))
        else:
            args.append((k, v))

    for k, v in args:
        v = v.annotation
        pre_sig.append(v if v is not inspect._empty else Any)

    for k, v in sorted(kwargs):
        v = v.annotation
        pre_sig.append(v if v is not inspect._empty else Any)

    v = sig.return_annotation
    pre_sig.append(v if v is not inspect._empty else Any)

    sig = inspect.signature(post)
    post_sig = []

    args = []
    kwargs = []

    for k, v in sig.parameters.items():
        if v.kind is v.VAR_KEYWORD or v.POSITIONAL_OR_KEYWORD:
            kwargs.append((k, v))
        else:
            args.append((k, v))

    for k, v in args:
        v = v.annotation
        post_sig.append(v if v is not inspect._empty else Any)

    for k, v in sorted(kwargs):
        v = v.annotation
        post_sig.append(v if v is not inspect._empty else Any)

    v = sig.return_annotation
    post_sig.append(v if v is not inspect._empty else Any)
    return all(_check(x, y) for x, y in zip_longest(pre_sig, post_sig, fillvalue=Any))


def _check(x, y):
    if config.debugging:
        print(
            f"checking:\n  "
            f"{x!r:<30} "
            f"({type(x).__module__}.{type(x).__qualname__})\n  "
            f"{y!r:<30} "
            f"({type(y).__module__}.{type(y).__qualname__})"
        )

    # narrowing {(Any => Any) | (Any => something)} is OK
    if x is Any:
        return True
    # broadening (something => Any) is NOK
    if y is Any:
        return False

    # now to the more specific cases (something => something else)
    with contextlib.suppress(TypeError):  # issubclass is allergic to container classes (types.GenericAlias)
        # let's first check if we're dealing with numbers
        if issubclass(x, Number) and issubclass(y, Number):
            # since the generic implementations aren't actual subclasses of each other
            # we have to map to the numeric tower classes
            for X in numerics:
                if issubclass(x, X):
                    break
            x = X
            for Y in numerics:
                if issubclass(y, Y):
                    break
            y = Y
    # Need to do this *before* we access `__name__`
    X = get_origin(x) or x
    Y = get_origin(y) or y

    # the other important type hierarchy are
    # containers
    # let's check if the user is using typing
    # classes and educate them to use
    # collections.abc instead
    for name, t in {"X": X, "Y": Y}.items():
        if t.__module__ != "typing":
            continue
        tca = getattr(collections.abc, t.__name__, 0)
        if not tca:
            continue
        warn(
            NotComparableWarning(
                f"{t} is of a type from the typing "
                " module, which can't be compared. "
                "Please use a type from the "
                "`collections.abc` module (`{tca!r}`)."
            )
        )
        locals().__setitem__(name, tca)

    if X is not None and Y is not None:
        # (Sequence => list) is OK, (list => Sequence) is NOK
        return (
            all(_check(x_, y_) for x_, y_ in zip_longest(get_args(x), get_args(y), fillvalue=Any))
            if issubclass(Y, X)
            else False
        )

    try:
        # (x => y) where y is not a subclass of x is NOK
        return issubclass(y, x)
    except TypeError:  # (int => list[int]) or (list[int] => int) NOK
        return False


# function to run code in a jupyter notebook sandbox
def run_in_notebook(code):
    """Run code in a jupyter notebook sandbox.

    This function is useful to run code in a jupyter notebook
    without having to run the notebook server.

    Parameters
    ----------
    code : str
        The code to run.

    Returns
    -------
    None

    """
    from IPython.display import Javascript, display

    display(Javascript(f"IPython.notebook.kernel.execute('{code}')"))



"""
This is where the story begins. Welcome to JustUse!
Only imports and project-global constants are defined here. 
All superfluous imports are deleted to clean up the namespace - and thus help()

"""


import os
import sys
import tempfile
from datetime import datetime, timezone
from enum import Enum, IntEnum
from logging import basicConfig, getLogger
from pathlib import Path
from uuid import uuid4
from warnings import catch_warnings, filterwarnings

from beartype import beartype
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning

sessionID = uuid4()
del uuid4

home = Path(os.getenv("JUSTUSE_HOME", str(Path.home() / ".justuse-python")))

try:
    home.mkdir(mode=0o755, parents=True, exist_ok=True)
except PermissionError:
    # this should fix the permission issues on android #80
    home = tempfile.mkdtemp(prefix="justuse_")


if sys.version_info < (3, 11):
    import tomli as toml
else:
    import tomllib as toml

(home / "config.toml").touch(mode=0o755, exist_ok=True)

with open(home / "config.toml", "rb") as f:
    config_dict = toml.load(f)
del toml

from use.pydantics import Configuration

config = Configuration(**config_dict)
del Configuration

config.logs.mkdir(mode=0o755, parents=True, exist_ok=True)
config.packages.mkdir(mode=0o755, parents=True, exist_ok=True)

for file in (
    config.logs / "usage.log",
    config.registry,
):
    file.touch(mode=0o755, exist_ok=True)


def fraction_of_day(now: datetime = None) -> float:
    if now is None:
        now = datetime.now(timezone.utc)
    return round(
        (
            now.hour / 24
            + now.minute / (24 * 60)
            + now.second / (24 * 60 * 60)
            + now.microsecond / (24 * 60 * 60 * 1000 * 1000)
        )
        * 1000,
        6,
    )


basicConfig(
    filename=config.logs / "usage.log",
    filemode="a",
    format="%(asctime)s %(levelname)s %(name)s %(message)s",
    datefmt=f"%Y%m%d {fraction_of_day()}",
    # datefmt="%Y-%m-%d %H:%M:%S",
    level=config.debug_level,
)

# !!! SEE NOTE !!!
# IMPORTANT; The setup.py script must be able to read the
# current use __version__ variable **AS A STRING LITERAL** from
# this file. If you do anything except updating the version,
# please check that setup.py can still be executed.
__version__ = "0.8.1.1"
# for tests
__version__ = os.getenv("USE_VERSION", __version__)
__name__ = "use"
__package__ = "use"

log = getLogger(__name__)
log.info(f"### SESSION START {datetime.now().strftime('%Y/%m/%d %H:%M:%S')} {sessionID} ###")

import use.logutil


class JustuseIssue:
    pass


class NirvanaWarning(Warning, JustuseIssue):
    pass


class VersionWarning(Warning, JustuseIssue):
    pass


class NotReloadableWarning(Warning, JustuseIssue):
    pass


class NoValidationWarning(Warning, JustuseIssue):
    pass


class AmbiguityWarning(Warning, JustuseIssue):
    pass


class UnexpectedHash(ImportError, JustuseIssue):
    pass


class InstallationError(ImportError, JustuseIssue):
    pass


# Coerce all PEP 585 deprecation warnings into fatal exceptions.
with catch_warnings():
    from beartype.roar import BeartypeDecorHintPep585DeprecationWarning

    filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning, module="beartype")


import hashlib


class Hash(Enum):
    sha256 = hashlib.sha256
    blake = hashlib.blake2s


del hashlib


class Modes(IntEnum):
    auto_install = 2**0
    fatal_exceptions = 2**1
    reloading = 2**2
    no_public_installation = 2**3
    fastfail = 2**4
    recklessness = 2**5
    no_browser = 2**6
    no_cleanup = 2**7


from use.aspectizing import (
    apply_aspect,
    iter_submodules,
    show_aspects,
    show_profiling,
    tinny_profiler,
    woody_logger,
)
from use.buffet import buffet_table
from use.main import URL, ProxyModule, Use, test_version
from use.pydantics import Version, git

for member in Modes:
    setattr(Use, member.name, member.value)

use = Use()

del os
use.__dict__.update(dict(globals()))
use = ProxyModule(use)
if not test_version:
    sys.modules["use"] = use

del test_version
del use.__dict__["test_version"]
del sys
del use.__dict__["sys"]

with catch_warnings():
    filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning, module="beartype")
    apply_aspect(use.iter_submodules(use), beartype)



"""
Pydantic model for the PyPI JSON API.
"""
import os
import re

from logging import INFO
from logging import getLogger
from pathlib import Path
from typing import Optional
from typing import Union

import packaging

from packaging.version import Version as PkgVersion
from pydantic import BaseModel

log = getLogger(__name__)


# Well, apparently they refuse to make Version iterable, so we'll have to do it ourselves.
# This is necessary to compare sys.version_info with Version and make some tests more elegant, amongst other things.


class Configuration(BaseModel):
    debug_level: int = INFO  # 20
    version_warning: bool = True
    no_browser: bool = False
    disable_jack: bool = bool(int(os.getenv("DISABLE_JACK", "0")))
    debugging: bool = bool(int(os.getenv("DEBUG", "0")))
    use_db: bool = bool(int(os.getenv("USE_DB", "1")))
    testing: bool = bool(int(os.getenv("TESTING", "0")))
    home: Path = Path(os.getenv("JUSTUSE_HOME", str(Path.home() / ".justuse-python"))).absolute()
    venv: Path = Path(os.getenv("JUSTUSE_HOME", str(Path.home() / ".justuse-python"))).absolute() / "venv"
    packages: Path = (
        Path(os.getenv("JUSTUSE_HOME", str(Path.home() / ".justuse-python"))).absolute() / "packages"
    )
    logs: Path = Path(os.getenv("JUSTUSE_HOME", str(Path.home() / ".justuse-python"))).absolute() / "logs"
    registry: Path = (
        Path(os.getenv("JUSTUSE_HOME", str(Path.home() / ".justuse-python"))).absolute() / "registry.db"
    )

    class Config:
        validate_assignment = True


class git(BaseModel):
    repo: str
    host: str = "github.com"
    branch: str = "main"
    commit: Optional[str] = None


class Version(PkgVersion):
    def __new__(cls, *args, **kwargs):
        if args and isinstance(args[0], Version):
            return args[0]
        else:
            return super(cls, Version).__new__(cls)

    def __init__(self, versionobj: Optional[Union[PkgVersion, str]] = None, *, major=0, minor=0, patch=0):
        if isinstance(versionobj, Version):
            return

        if versionobj:
            super(Version, self).__init__(versionobj)
            return

        if major is None or minor is None or patch is None:
            raise ValueError(
                f"Either 'Version' must be initialized with either a string, packaging.version.Verson, {__class__.__qualname__}, or else keyword arguments for 'major', 'minor' and 'patch' must be provided. Actual invocation was: {__class__.__qualname__}({versionobj!r}, {major=!r}, {minor=!r})"
            )

        # string as only argument
        # no way to construct a Version otherwise - WTF
        versionobj = ".".join(map(str, (major, minor, patch)))
        super(Version, self).__init__(versionobj)

    def __iter__(self):
        yield from self.release

    def __repr__(self):
        return f"Version('{super().__str__()}')"

    def __hash__(self):
        return hash(self._version)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        return Version(value)


def _delete_none(a_dict: dict[str, object]) -> dict[str, object]:
    for k, v in tuple(a_dict.items()):
        if v is None or v == "":
            del a_dict[k]
    return a_dict


class RegistryEntry(BaseModel):
    class Config:
        validate_assignment = True

    artifact_path: Path
    installation_path: Path
    pure_python_package: bool


class JustUse_Info(BaseModel):
    distribution: Optional[str]
    version: Optional[str]
    build_tag: Optional[str]
    python_tag: Optional[str]
    abi_tag: Optional[str]
    platform_tag: Optional[str]
    ext: Optional[str]


class PyPI_Release(BaseModel):
    abi_tag: Optional[str]
    build_tag: Optional[str]
    distribution: Optional[str]
    digests: dict[str, str]
    ext: Optional[str]
    filename: str
    requires_python: Optional[str]
    packagetype: str
    platform_tag: Optional[str]
    python_version: Optional[str]
    python_tag: Optional[str]
    url: str
    version: Version
    yanked: bool

    class Config:
        validate_assignment = True

    @property
    def is_sdist(self):
        return (
            self.packagetype == "sdist" or self.python_version == "source" or self.justuse.abi_tag == "none"
        )

    @property
    def justuse(self) -> JustUse_Info:
        pp = Path(self.filename)
        if ".tar" in self.filename:
            ext = self.filename[self.filename.index(".tar") + 1 :]
        else:
            ext = pp.name[len(pp.stem) + 1 :]
        rest = pp.name[: -len(ext) - 1]

        not_dash = lambda name: f"(?P<{name}>[^-]+)"
        not_dash_with_int = lambda name: f"(?P<{name}>[0-9][^-]*)"
        if match := re.match(
            f"{not_dash('distribution')}-{not_dash('version')}-?{not_dash_with_int('build_tag')}?-?{not_dash('python_tag')}?-?{not_dash('abi_tag')}?-?{not_dash('platform_tag')}?",
            rest,
        ):
            return JustUse_Info(**_delete_none(match.groupdict()), ext=ext)
        return JustUse_Info()


class PyPI_Downloads(BaseModel):
    last_day: int
    last_month: int
    last_week: int


class PyPI_Info(BaseModel):
    class Config:
        extra = "ignore"

    description_content_type: Optional[str]
    download_url: Optional[str]
    package_name: Optional[str]
    package_url: str
    platform: Optional[str]
    project_url: Optional[str]
    project_urls: Optional[dict[str, str]]
    release_url: Optional[str]
    requires_dist: Optional[list[str]]
    requires_python: Optional[str]
    summary: Optional[str]
    version: Optional[str]
    yanked: Optional[bool]
    yanked_reason: Optional[str]


class PyPI_URL(BaseModel):
    abi_tag: Optional[str]
    build_tag: Optional[str]
    digests: dict[str, str]
    url: str
    packagetype: str
    requires_python: Optional[str]
    python_version: Optional[str]
    filename: str
    yanked: bool
    distribution: Optional[str]
    python_tag: Optional[str]
    platform_tag: Optional[str]
    ext: Optional[str]


class PyPI_Project(BaseModel):
    releases: Optional[dict[Version, list[PyPI_Release]]] = {}
    urls: list[PyPI_URL] = None
    last_serial: int = None
    info: PyPI_Info = None

    class Config:
        extra = "ignore"

    def __init__(self, *, releases=None, urls, info, **kwargs):
        try:
            for version in list(releases.keys()):
                if not isinstance(version, str):
                    continue
                try:
                    Version(version)
                except packaging.version.InvalidVersion:
                    del releases[version]

            def get_info(rel_info, ver_str):
                data = {
                    **rel_info,
                    **_parse_filename(rel_info["filename"]),
                    "version": Version(str(ver_str)),
                }
                if info.get("requires_python"):
                    data["requires_python"] = info.get("requites_python")
                if info.get("requires_dist"):
                    data["requires_dist"] = info.get("requires_dist")
                return data

            super(PyPI_Project, self).__init__(
                releases={
                    str(ver_str): [get_info(rel_info, ver_str) for rel_info in release_infos]
                    for ver_str, release_infos in releases.items()
                },
                urls=[
                    get_info(rel_info, ver_str)
                    for ver_str, rel_infos in releases.items()
                    for rel_info in rel_infos
                ],
                info=info,
                **kwargs,
            )
        except AttributeError:
            pass
        finally:
            releases = None
            info = None
            urls = None


def _parse_filename(filename) -> dict:
    """
    REFERENCE IMPLEMENTATION - DO NOT USE
    Match the filename and return a dict of parts.
    >>> parse_filename("numpy-1.19.5-cp36-cp36m-macosx_10_9_x86_64.whl")
    {'distribution': 'numpy', 'version': '1.19.5', 'build_tag', 'python_tag': 'cp36', 'abi_tag': 'cp36m', 'platform_tag': 'macosx_10_9_x86_64', 'ext': 'whl'}
    """
    # Filename as API, seriously WTF...
    assert isinstance(filename, str)
    distribution = version = build_tag = python_tag = abi_tag = platform_tag = None
    pp = Path(filename)
    packagetype = None
    if ".tar" in filename:
        ext = filename[filename.index(".tar") :]
        packagetype = "source"
    else:
        ext = pp.name[len(pp.stem) + 1 :]
        packagetype = "bdist"
    rest = pp.name[: -len(ext) - 1]

    p = rest.split("-")
    np = len(p)
    if np == 2:
        distribution, version = p
    elif np == 3:
        distribution, version, python_tag = p
    elif np == 5:
        distribution, version, python_tag, abi_tag, platform_tag = p
    elif np == 6:
        distribution, version, build_tag, python_tag, abi_tag, platform_tag = p
    else:
        return {}

    return {
        "distribution": distribution,
        "version": version,
        "build_tag": build_tag,
        "python_tag": python_tag,
        "abi_tag": abi_tag,
        "platform_tag": platform_tag,
        "ext": ext,
        "filename": filename,
        "packagetype": packagetype,
        "yanked_reason": "",
        "bugtrack_url": "",
    }



import os
import pathlib
import shutil
import sys

from colorama import Fore


def get_sources_and_destination() -> tuple[list[str], str]:
    """Get the sources and destination from :code:`sys.argv`"""
    return sys.argv[1:-1], sys.argv[-1]


def log_move(source_path: str, source_to_destination_path: str) -> None:
    """Log when files are moved."""
    print(f"{Fore.YELLOW}{source_path} {Fore.RESET}→ {Fore.GREEN}{source_to_destination_path}{Fore.RESET}")


def log_not_found(path: str) -> None:
    """Log when a path/file doesn't exist."""
    print(f"{Fore.RED}Could not find {Fore.YELLOW}{path}{Fore.RESET}.")


def main() -> None:
    if len(sys.argv) < 3:
        return

    sources, destination = get_sources_and_destination()

    for source in sources:
        if not os.path.exists(source):
            log_not_found(source)
            continue

        log_move(str(pathlib.Path(source)), shutil.move(source, destination))


if __name__ == "__main__":
    main()



# here we're building the buffet of the future with pattern matching (>=3.10)

from logging import getLogger

log = getLogger(__name__)

from use.messages import UserMessage as Message
from use.pimp import (
    _auto_install,
    _ensure_version,
    _import_public_no_install,
    _pebkac_no_hash,
    _pebkac_no_version,
    _pebkac_no_version_no_hash,
)
from use.tools import pipes


# fmt: off
@pipes
def buffet_table(case, kwargs):
    match case: # type: ignore
#            +-------------------------- version specified?
#            |  +----------------------- hash specified?
#            |  |  +-------------------- already publicly available?
#            |  |  |  +----------------- auto-install requested?
#            |  |  |  |
#            v  v  v  v      
        case _, _, 0, 0: return ImportError(Message.cant_import(kwargs["package_name"]))
        case 0, _, 1, 0: return _import_public_no_install(**kwargs)
        case 1, _, 1, 0: return _import_public_no_install(**kwargs) >> _ensure_version(**kwargs)
        case 0, 0, _, 1: return _pebkac_no_version_no_hash(**kwargs)
        case 1, 0, _, 1: return _pebkac_no_hash(**kwargs)
        case 0, 1, _, 1: return _pebkac_no_version(**kwargs)
        case 1, 1, 0, 1: return _auto_install(**kwargs)
        case 1, 1, 1, 1: return _import_public_no_install(**kwargs) >> _ensure_version(**kwargs) >> _auto_install(**kwargs)
# fmt: on



import ast
import contextlib
import inspect
import re
import sys
from collections import namedtuple
from collections.abc import Callable, Iterable
from enum import Enum
from functools import wraps
from logging import getLogger
from pathlib import Path
from time import perf_counter_ns
from types import ModuleType
from typing import Any, DefaultDict, Deque, Optional, Union

from beartype import beartype
from icontract import require

log = getLogger(__name__)

from use import config
from use.messages import _web_aspectized, _web_aspectized_dry_run, _web_tinny_profiler

# TODO: use an extra WeakKeyDict as watchdog for object deletions and trigger cleanup in these here
_applied_decorators: DefaultDict[tuple[object, str], Deque[Callable]] = DefaultDict(Deque)
"to see which decorators are applied, in which order"
_aspectized_functions: DefaultDict[tuple[object, str], Deque[Callable]] = DefaultDict(Deque)
"the actually decorated functions to undo aspectizing"


def show_aspects():
    """Open a browser to properly display all the things that have been aspectized thus far."""
    print("decorators:", _applied_decorators)
    print("functions:", _aspectized_functions)
    # _web_aspectized(_applied_decorators, _aspectized_functions)


def is_callable(thing):
    try:
        return callable(
            object.__getattribute__(type(thing), "__call__")
        )  # to even catch weird non-callable __call__ methods
    except AttributeError:
        return False


HIT = namedtuple("Hit", "qualname name type success exception dunder module_name")


def really_callable(thing):
    try:
        thing.__get__
        thing.__call__
        return True
    except Exception:
        return False


def apply_aspect(
    thing: Union[object, Iterable[object]],
    decorator: Callable,
    /,
    *,
    check: Callable = is_callable,
    dry_run: bool = False,
    pattern: str = "",
    excluded_names: Optional[set[str]] = None,
    excluded_types: Optional[set[type]] = None,
    file=None,
) -> None:
    """Apply the aspect as a side-effect, no copy is created."""

    regex = re.compile(pattern, re.DOTALL)

    if excluded_names is None:
        excluded_names = set()
    if excluded_types is None:
        excluded_types = set()

    visited = {id(obj) for obj in vars(object).values()}
    visited.add(id(type))

    hits = []

    def aspectize(
        thing: Any,
        decorator: Callable,
        /,
        *,
        qualname_lst: Optional[list] = None,
        module_name: str,
    ) -> Iterable[HIT]:
        name = getattr(thing, "__name__", str(thing))
        if not qualname_lst:
            qualname_lst = []
        if id(thing) in visited:
            return

        qualname_lst.append(name)

        # let's stick within the module boundary
        try:
            thing_module_name = inspect.getmodule(thing).__name__
        except AttributeError:
            thing_module_name = "<unknown>"
        if thing_module_name != module_name:
            return

        for name in dir(thing):
            qualname = "" if len(qualname_lst) < 3 else "..." + ".".join(qualname_lst[-3:]) + "." + name
            obj = getattr(thing, name, None)
            qualname = getattr(obj, "__qualname__", qualname)

            if obj is None:
                continue
            # Time to get serious!

            if type(obj) == type:
                aspectize(obj, decorator, qualname_lst=qualname_lst, module_name=module_name)

            if id(obj) in visited:
                continue

            if (
                qualname in excluded_names
                or type(obj) in excluded_types
                or not check(obj)
                or not regex.match(name)
            ):
                continue

            msg = ""
            success = True

            try:
                wrapped = _wrap(thing=thing, obj=obj, decorator=decorator, name=name)
                if dry_run:
                    _unwrap(thing=thing, name=name)
            except BaseException as exc:
                wrapped = obj
                success = False
                msg = str(exc)
                if file:
                    print(msg, file=file)
            assert isinstance(name, str) and len(name) > 0
            assert isinstance(module_name, str) and module_name != ""

            hits.append(
                HIT(
                    qualname,
                    name,
                    type(wrapped),
                    success,
                    msg,
                    name.startswith("__") and name.endswith("__"),
                    module_name,
                )
            )
            visited.add(id(wrapped))

    def call(m):
        try:
            module_name = inspect.getmodule(m).__name__
        except AttributeError:
            module_name = "<unknown>"
        aspectize(m, decorator, module_name=module_name)

    if isinstance(thing, Iterable):
        for x in thing:
            call(x)
    else:
        call(thing)

    if dry_run:
        if config.no_browser:
            print("Tried to do a dry run and display the results, but no_browser is set in config.")
        else:
            print("Please check your browser to select options and filters for aspects.")
            _web_aspectized_dry_run(
                decorator=decorator, pattern=pattern, check=check, hits=hits, module_name=str(thing)
            )


@beartype
def _wrap(*, thing: Any, obj: Any, decorator: Callable, name: str) -> Any:
    wrapped = decorator(obj)
    _applied_decorators[(thing, name)].append(decorator)
    _aspectized_functions[(thing, name)].append(obj)

    # This will fail with TypeError on built-in/extension types.
    # We handle exceptions outside, let's not betray ourselves.
    setattr(thing, name, wrapped)
    return wrapped


@beartype
def _unwrap(*, thing: Any, name: str):
    try:
        original = _aspectized_functions[(thing, name)].pop()
    except IndexError:
        del _aspectized_functions[(thing, name)]
        original = getattr(thing, name)

    try:
        _applied_decorators[(thing, name)].pop()
    except IndexError:
        del _applied_decorators[(thing, name)]
    setattr(thing, name, original)

    return original


def _qualname(thing):
    module = getattr(thing, "__module__", None) or getattr(thing.__class__, "__module__", None)
    qualname = getattr(thing, "__qualname__", None) or getattr(thing, "__name__", None) or str(thing)
    return f"{module}::{qualname}"


def woody_logger(thing: Callable) -> Callable:
    """
    Decorator to log/track/debug calls and results.

    Args:
        func (function): The function to decorate.
    Returns:
        function: The decorated callable.
    """
    # A class is an instance of type - its type is also type, but only checking for `type(thing) is type`
    # would exclude metaclasses other than type - not complicated at all.
    if isinstance(thing, type):
        # wrapping the class means wrapping `__new__` mainly, which must return the original class, not just a proxy -
        # because a proxy never could hold up under scrutiny of type(), which can't be faked (like, at all).
        class wrapper(thing.__class__):
            def __new__(cls, *args, **kwargs):
                print(f"{args} {kwargs} -> {thing.__name__}()")
                before = perf_counter_ns()
                res = thing(*args, **kwargs)
                after = perf_counter_ns()
                print(
                    f"-> {thing.__name__}() (in {after - before} ns ({round((after - before) / 10**9, 5)} sec) -> {type(res)}"
                )
                return res

    else:
        # Wrapping a function is way easier..
        @wraps(thing)
        def wrapper(*args, **kwargs):
            print(f"{args} {kwargs} -> {_qualname(thing)}")
            before = perf_counter_ns()
            res = thing(*args, **kwargs)
            after = perf_counter_ns()
            print(
                f"-> {_qualname(thing)} (in {after - before} ns ({round((after - before) / 10**9, 5)} sec) -> {res} {type(res)}"
            )
            return res

    return wrapper


_timings: dict[int, Deque[int]] = DefaultDict(lambda: Deque(maxlen=10000))


def tinny_profiler(func: callable) -> callable:
    """
    Decorator to log/track/debug calls and results.

    Args:
        func (function): The function to decorate.
    Returns:
        function: The decorated callable.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        before = perf_counter_ns()
        res = func(*args, **kwargs)
        after = perf_counter_ns()
        _timings[func].append(after - before)
        return res

    return wrapper


def show_profiling():
    _web_tinny_profiler(_timings)


def _is_builtin(name, mod):
    if name in sys.builtin_module_names:
        return True

    if hasattr(mod, "__file__"):
        relpath = Path(mod.__file__).parent.relative_to((Path(sys.executable).parent / "lib"))
        if relpath == Path():
            return True
        if relpath.parts[0] == "site-packages":
            return False
    return True


def _get_imports_from_module(mod):
    if not hasattr(mod, "__file__"):
        return
    with open(mod.__file__, "rb") as file:
        with contextlib.suppress(ValueError):
            for x in ast.walk(ast.parse(file.read())):
                if isinstance(x, ast.Import):
                    name = x.names[0].name
                    if (mod := sys.modules.get(name)) and not _is_builtin(name, mod):
                        yield name, mod
                if isinstance(x, ast.ImportFrom):
                    name = x.module
                    if (mod := sys.modules.get(name)) and not _is_builtin(name, mod):
                        yield name, mod


def iter_submodules(mod: ModuleType, visited=None, results=None) -> set[ModuleType]:
    """Find all modules recursively that were imported as dependency from  the given module."""
    if results is None:
        results = set()
    if visited is None:
        visited = set()
    for name, x in _get_imports_from_module(mod):
        if name in visited:
            continue
        visited.add(name)
        results.add(name)
        for name, x in _get_imports_from_module(sys.modules[name]):
            results.add(x)
            iter_submodules(x, visited, results)
    return results



import io
import logging
import os
import sys
import time
import traceback
from logging import Formatter, PercentStyle, StreamHandler, StrFormatStyle, StringTemplateStyle
from typing import NamedTuple, Deque
from collections.abc import Callable

import use

BASIC_FORMAT: str = "%(levelname)s:%(name)s:%(message)s"

_STYLES: dict[str, tuple[PercentStyle, str]] = {
    "%": (PercentStyle, BASIC_FORMAT),
    "{": (StrFormatStyle, "{levelname}:{name}:{message}"),
    "$": (
        StringTemplateStyle,
        "\x1b[1;3${levelno}m${levelname}\x1b[1;30m: \x1b[0m\x1b[1;36m${name}\x1b[1;30m: \x1b[0m${message}",
    ),
}


class TimeResult(
    NamedTuple(
        "TimeResult",
        tm_year=int,
        tm_mon=int,
        tm_mday=int,
        tm_hour=int,
        tm_min=int,
        tm_sec=int,
        tm_wday=int,
        tm_yday=int,
        tm_isdst=bool,
    )
):
    pass


class ConsoleFormatter(Formatter):
    """
    Formatter instances are used to convert a LogRecord to text.

    Formatters need to know how a LogRecord is constructed. They are
    responsible for converting a LogRecord to (usually) a string which can
    be interpreted by either a human or an external system. The base Formatter
    allows a formatting string to be specified. If none is supplied, the
    style-dependent default value, "%(message)s", "{message}", or
    "${message}", is used.

    The Formatter can be initialized with a format string which makes use of
    knowledge of the LogRecord attributes - e.g. the default value mentioned
    above makes use of the fact that the user's message and arguments are pre-
    formatted into a LogRecord's message attribute. Currently, the useful
    attributes in a LogRecord are described by:

    %(name)s            Name of the logger (logging channel)
    %(levelno)s         Numeric logging level for the message (DEBUG, INFO,
                        WARNING, ERROR, CRITICAL)
    %(levelname)s       Text logging level for the message ("DEBUG", "INFO",
                        "WARNING", "ERROR", "CRITICAL")
    %(pathname)s        Full pathname of the source file where the logging
                        call was issued (if available)
    %(filename)s        Filename portion of pathname
    %(module)s          Module (name portion of filename)
    %(lineno)d          Source line number where the logging call was issued
                        (if available)
    %(funcName)s        Function name
    %(created)f         Time when the LogRecord was created (time.time()
                        return value)
    %(asctime)s         Textual time when the LogRecord was created
    %(msecs)d           Millisecond portion of the creation time
    %(relativeCreated)d Time in milliseconds when the LogRecord was created,
                        relative to the time the logging module was loaded
                        (typically at application startup time)
    %(thread)d          Thread ID (if available)
    %(threadName)s      Thread name (if available)
    %(process)d         Process ID (if available)
    %(message)s         The result of record.getMessage(), computed just as
                        the record is emitted
    """

    converter: Callable[[int], TimeResult]
    style: tuple[PercentStyle, str]
    default_time_format: str = "%Y-%m-%d %H:%M:%S"
    default_msec_format: str = "%s,%03d"
    level = property(lambda self: logging.root.level)

    def __init__(self):
        """
        Initialize the formatter
        """
        self.datefmt = "%Y-%m-%d"
        style = "$"
        fmt = _STYLES[style][1]
        super(ConsoleFormatter, self).__init__(fmt=fmt, datefmt=self.datefmt, style=style, validate=True)

    def formatTime(self, record, datefmt=None):
        """
        Return the creation time of the specified LogRecord as formatted text.

        This method should be called from format() by a formatter which
        wants to make use of a formatted time. This method can be overridden
        in formatters to provide for any specific requirement, but the
        basic behaviour is as follows: if datefmt (a string) is specified,
        it is used with time.strftime() to format the creation time of the
        record. Otherwise, an ISO8601-like (or RFC 3339-like) format is used.
        The resulting string is returned. This function uses a user-configurable
        function to convert the creation time to a tuple. By default,
        time.localtime() is used; to change this for a particular formatter
        instance, set the 'converter' attribute to a function with the same
        signature as time.localtime() or time.gmtime(). To change it for all
        formatters, for example if you want all logging times to be shown in GMT,
        set the 'converter' attribute in the Formatter class.
        """
        if datefmt is None:
            datefmt = self.datefmt
        ct = self.converter(record.created)
        if datefmt:
            s = time.strftime(datefmt, ct)
        else:
            s = time.strftime(self.default_time_format, ct)
            if self.default_msec_format:
                s = self.default_msec_format % (s, record.msecs)
        return s

    def formatException(self, ei):
        """
        Format and return the specified exception information as a string.

        This default implementation just uses
        traceback.print_exception()
        """
        sio = io.StringIO()
        tb = ei[2]
        # See issues #9427, #1553375. Commented out for now.
        # if getattr(self, 'fullstack', False):
        #    traceback.print_stack(tb.tb_frame.f_back, file=sio)
        traceback.print_exception(ei[0], ei[1], tb, None, sio)
        s = sio.getvalue()
        sio.close()
        if s[-1:] == "\n":
            s = s[:-1]
        return s

    def usesTime(self):
        """
        Check if the format uses the creation time of the record.
        """
        return True

    def formatMessage(self, record):
        return self._style.format(record)

    def formatStack(self, stack_info):
        """
        This method is provided as an extension point for specialized
        formatting of stack information.

        The input data is a string as returned from a call to
        :func:`traceback.print_stack`, but with the last trailing newline
        removed.

        The base implementation just returns the value passed in.
        """
        return stack_info

    def format(self, record):
        """
        Format the specified record as text.

        The record's attribute dictionary is used as the operand to a
        string formatting operation which yields the returned string.
        Before formatting the dictionary, a couple of preparatory steps
        are carried out. The message attribute of the record is computed
        using LogRecord.getMessage(). If the formatting string uses the
        time (as determined by a call to usesTime(), formatTime() is
        called to format the event time. If there is exception information,
        it is formatted using formatException() and appended to the message.
        """
        record.message = record.getMessage()
        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)
        s = self.formatMessage(record)
        if record.exc_info and not record.exc_text:
            record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + record.exc_text
        if record.stack_info:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + self.formatStack(record.stack_info)
        return s


class ConsoleHandler(StreamHandler):
    """
    A handler class which writes logging records, appropriately formatted,
    to a stream. Note that this class does not close the stream, as
    sys.stdout or sys.stderr may be used.
    """

    terminator: str
    stream: io.IOBase
    formatter: ConsoleFormatter

    def __init__(self):
        """
        Initialize the handler.
        """
        self.level = logging.root.level
        formatter = ConsoleFormatter()
        self.formatter = formatter
        self.fmt = formatter
        super(ConsoleHandler, self).__init__(stream=sys.stderr)
        super(StreamHandler, self).__init__(level=logging.DEBUG)
        self.formatter = formatter
        self.fmt = formatter

    def flush(self):
        """
        Flushes the stream.
        """
        self.acquire()
        try:
            if self.stream and hasattr(self.stream, "flush"):
                self.stream.flush()
        finally:
            self.release()

    def format(self, record):
        """
        Format the specified record.

        If a formatter is set, use it. Otherwise, use the default formatter
        for the module.
        """
        if self.formatter is None:
            self.formatter = ConsoleFormatter()
        fmt: ConsoleFormatter = self.formatter
        return fmt.format(record)

    def emit(self, record):
        """
        Emit a record.

        If a formatter is specified, it is used to format the record.
        The record is then written to the stream with a trailing newline.  If
        exception information is present, it is formatted using
        traceback.print_exception and appended to the stream.  If the stream
        has an 'encoding' attribute, it is used to determine how to do the
        output to the stream.
        """
        try:
            msg = self.format(record)
            stream = self.stream
            # issue 35046: merged two stream.writes into one.
            stream.write(msg + self.terminator)
            self.flush()
        except (RecursionError, Exception):  # See issue 36272
            raise

    def setStream(self, stream):
        """
        Sets the StreamHandler's stream to the specified value,
        if it is different.

        Returns the old stream, if the stream was changed, or None
        if it wasn't.
        """
        if stream is self.stream:
            result = None
        else:
            result = self.stream
            self.acquire()
            try:
                self.flush()
                self.stream = stream
            finally:
                self.release()
        return result

    def __repr__(self):
        return f"<{self.__class__.__name__}>"


if use.config.debugging:
    logging.root.setLevel(logging.DEBUG)
handler = ConsoleHandler()
logging.root.handlers.append(handler)



import os
from pathlib import Path

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))
src = os.path.join(here, "src/use")

# Instead of doing the obvious thing (importing 'use' directly and just reading '__version__'),
# we are parsing the version out of the source AST here, because if the user is missing any
# dependencies at setup time, an import error would prevent the installation.
# Two simple ways to verify the installation using this setup.py file:
#
#     python3 setup.py develop
#
#  or:
#
#    python3 setup.py install
#
import ast

with open(os.path.join(src, "__init__.py")) as f:
    mod = ast.parse(f.read())
    version = [
        t
        for t in [*filter(lambda n: isinstance(n, ast.Assign), mod.body)]
        if t.targets[0].id == "__version__"
    ][0].value.value

meta = {
    "name": "justuse",
    "license": "MIT",
    "url": "https://github.com/amogorkon/justuse",
    "version": version,
    "author": "Anselm Kiefner",
    "author_email": "justuse-pypi@anselm.kiefner.de",
    "python_requires": ">=3.10",
    "keywords": [
        "installing",
        "packages",
        "hot reload",
        "auto install",
        "aspect oriented",
        "version checking",
        "functional",
    ],
    "classifiers": [
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
    ],
    "extras_require": {"test": ["pytest", "pytest-cov", "pytest-env"]},
    "fullname": "justuse",
    "dist_files": ["pytest.ini", "tests/pytest.ini"],
    "description": "a pure-python alternative to import",
    "maintainer_email": "justuse-pypi@anselm.kiefner.de",
    "maintainer": "Anselm Kiefner",
    "platforms": ["any"],
    "download_url": "https://github.com/amogorkon/justuse/" "archive/refs/heads/main.zip",
}


requires = (
    "beartype( >= 0.8.1)",
    "furl(>= 2.1.2)",
    "icontract(>= 2.5.4)",
    "jinja2",
    "packaging(== 21.0)",
    "pip",
    "pydantic(>= 1.8.2)",
    "requests",
    "wheel",
)


LONG_DESCRIPTION = Path("README.md").read_text()
setup(
    packages=find_packages(where="src"),
    package_dir={
        "": "src",
    },
    package_name="use",
    include_package_data=True,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    requires=requires,
    install_requires=requires,
    setup_requires=requires,
    zip_safe=True,
    **meta
)



"""
Main classes that act as API for the user to interact with.
"""



import asyncio
import atexit
import hashlib
import importlib.util
import inspect
import os
import shutil
import sqlite3
import sys
import threading
import time
import traceback
from collections import namedtuple
from collections.abc import Callable
from datetime import datetime
from logging import DEBUG, INFO, NOTSET, getLogger, root
from pathlib import Path
from types import ModuleType
from typing import Any, Optional, Union
from warnings import warn

import requests
from furl import furl as URL
from icontract import require

from use import (
    AmbiguityWarning,
    Hash,
    Modes,
    NotReloadableWarning,
    NoValidationWarning,
    UnexpectedHash,
    VersionWarning,
    __version__,
    buffet_table,
    config,
    home,
    sessionID,
)
from use.aspectizing import _applied_decorators, apply_aspect
from use.hash_alphabet import JACK_as_num, is_JACK, num_as_hexdigest
from use.messages import KwargMessage, StrMessage, TupleMessage, UserMessage
from use.pimp import (
    _build_mod,
    _ensure_path,
    _fail_or_default,
    _modules_are_compatible,
    _parse_name,
    _real_path,
)
from use.pydantics import Version, git
from use.tools import methdispatch

log = getLogger(__name__)

# internal subpackage imports
test_version: str = locals().get("test_version")

_reloaders: dict["ProxyModule", "ModuleReloader"] = {}  # ProxyModule:Reloader
_using = {}

# sometimes all you need is a sledge hammer..
def _release_locks():
    for _ in range(2):
        [lock.unlock() for lock in threading._shutdown_locks]
        [reloader.stop() for reloader in _reloaders.values()]
    log.info(f"### SESSION END {datetime.now().strftime('%Y/%m/%d %H:%M:%S')} {sessionID} ###")


atexit.register(_release_locks)

ModInUse = namedtuple("ModInUse", "name mod path spec")


class ProxyModule(ModuleType):
    def __init__(self, mod):
        self.__implementation = mod
        self.__condition = threading.RLock()

    def __getattribute__(self, name):
        if name in (
            "_ProxyModule__implementation",
            "_ProxyModule__condition",
            "",
            "__class__",
            "__metaclass__",
            "__instancecheck__",
        ):
            return object.__getattribute__(self, name)
        with self.__condition:
            return getattr(self.__implementation, name)

    def __setattr__(self, name, value):
        if name in (
            "_ProxyModule__implementation",
            "_ProxyModule__condition",
        ):
            object.__setattr__(self, name, value)
            return
        with self.__condition:
            setattr(self.__implementation, name, value)

    def __matmul__(self, other: Callable):
        thing = self.__implementation
        if not other:
            raise NotImplementedError

        # a little hack in order to be able to do `use @ numpy`...
        if isinstance(self.__implementation, Use):
            thing = other

            def some_decorator(x):
                return x

            other = some_decorator

        assert isinstance(other, Callable)

        kwargs = {
            "excluded_types": {
                ProxyModule,
            },
            "dry_run": True,
        }

        return apply_aspect(thing, other, **kwargs)

    def __call__(self, *args, **kwargs):
        with self.__condition:
            return self.__implementation(*args, **kwargs)

    # to allow `numpy @ use` for a quick check
    def __rmatmul__(self, *args, **kwargs):
        return ProxyModule.__matmul__(self, *args, **kwargs)


class ModuleReloader:
    def __init__(self, *, proxy, name, path, package_name, initial_globals):
        self.proxy = proxy
        "ProxyModula that we refer to."
        self.name = name
        self.path = path
        self.package_name = package_name
        self.initial_globals = initial_globals
        self._condition = threading.RLock()
        self._stopped = True
        self._thread = None

    def start_async(self):
        loop = asyncio.get_running_loop()
        loop.create_task(self.run_async())

    @require(lambda self: self._thread is None or self._thread.is_alive())
    def start_threaded(self):
        self._stopped = False
        atexit.register(self.stop)
        self._thread = threading.Thread(target=self.run_threaded, name=f"reloader__{self.name}")
        self._thread.start()

    async def run_async(self):
        last_filehash = None
        while not self._stopped:
            with open(self.path, "rb") as file:
                code = file.read()
            current_filehash = hashlib.blake2b(code).hexdigest()
            if current_filehash != last_filehash:
                try:
                    mod = _build_mod(
                        module_name=self.name,
                        code=code,
                        initial_globals=self.initial_globals,
                        module_path=self.path.resolve(),
                    )
                    if not _modules_are_compatible(self.proxy, mod):
                        continue
                    self.proxy.__implementation = mod
                except KeyError:
                    traceback.print_exc()
            last_filehash = current_filehash
            await asyncio.sleep(1)

    def run_threaded(self):
        last_filehash = None
        while not self._stopped:
            with self._condition:
                with open(self.path, "rb") as file:
                    code = file.read()
                current_filehash = hashlib.blake2b(code).hexdigest()
                if current_filehash != last_filehash:
                    try:
                        mod = _build_mod(
                            module_name=self.name,
                            code=code,
                            initial_globals=self.initial_globals,
                            module_path=self.path,
                        )
                        if not _modules_are_compatible(self.proxy, mod):
                            continue
                        self.proxy._ProxyModule__implementation = mod
                    except KeyError:
                        traceback.print_exc()
                last_filehash = current_filehash
            time.sleep(1)

    def stop(self):
        self._stopped = True

    def __del__(self):
        self.stop()


class Use(ModuleType):
    """
    # Welcome to the world of use

    *asdf* adsf
    """

    def __init__(self):
        self._using = _using
        # might run into issues during testing otherwise
        self.registry = self._set_up_registry()
        "Registry sqlite DB to store all relevant package metadata."

        if config.debugging:
            root.setLevel(DEBUG)

        if config.version_warning:
            try:
                response = requests.get("https://pypi.org/pypi/justuse/json")
                "Checking if there's a new version of justuse."
                data = response.json()
                max_version = max(Version(version) for version in data["releases"].keys())
                if Version(__version__) < max_version:
                    warn(
                        UserMessage.use_version_warning(max_version),
                        VersionWarning,
                    )
            except (KeyError, requests.exceptions.ConnectionError):
                if test_version:
                    raise
                log.error(
                    traceback.format_exc()
                )  # we really don't need to bug the user about this (either pypi is down or internet is broken)

    def clean_slate(self):
        shutil.rmtree(config.venv, ignore_errors=True)
        shutil.rmtree(config.packages, ignore_errors=True)
        config.venv.mkdir(mode=0o755, exist_ok=True)
        config.packages.mkdir(mode=0o755, parents=True, exist_ok=True)
        self.recreate_registry()

    def _set_up_registry(self, *, registry=None, path: Optional[Path] = None):
        # recreating reuses the registry connection and file
        if registry is None:
            if path or test_version and "DB_TEST" not in os.environ:
                registry = sqlite3.connect(path or ":memory:").cursor()
            else:
                try:
                    registry = sqlite3.connect(home / "registry.db").cursor()
                except Exception as e:
                    raise RuntimeError(UserMessage.couldnt_connect_to_db(e)) from e
        registry.row_factory = lambda cursor, row: {
            col[0]: row[idx] for idx, col in enumerate(cursor.description)
        }
        registry.execute("PRAGMA foreign_keys=ON")
        registry.execute("PRAGMA auto_vacuum = FULL")
        registry.executescript(
            """
CREATE TABLE IF NOT EXISTS "artifacts" (
	"id"    INTEGER,
	"distribution_id"   INTEGER,
	"import_relpath" TEXT,
	"artifact_path" TEXT,
  "module_path" TEXT,
	PRIMARY KEY("id" AUTOINCREMENT),
	FOREIGN KEY("distribution_id") REFERENCES "distributions"("id") ON DELETE CASCADE
);

CREATE TABLE IF NOT EXISTS "distributions" (
	"id"    INTEGER,
	"name"  TEXT NOT NULL,
	"version"   TEXT NOT NULL,
	"installation_path" TEXT,
	"date_of_installation"  INTEGER,
	"number_of_uses"    INTEGER,
	"date_of_last_use"  INTEGER,
	"pure_python_package"   INTEGER NOT NULL DEFAULT 1,
	PRIMARY KEY("id" AUTOINCREMENT)
);

CREATE TABLE IF NOT EXISTS "hashes" (
	"algo"  TEXT NOT NULL,
	"value" INTEGER NOT NULL,
	"artifact_id"   INTEGER NOT NULL,
	PRIMARY KEY("algo","value"),
	FOREIGN KEY("artifact_id") REFERENCES "artifacts"("id") ON DELETE CASCADE
);
		"""
        )
        registry.connection.commit()
        return registry

    def recreate_registry(self):
        number_of_backups = len(list(home.glob("registry.db*")))
        shutil.copyfile(home / "registry.db", home / f"registry.db.{number_of_backups}.bak")
        self._clear_registry()
        self._set_up_registry(registry=self.registry)
        self.cleanup()

    def _clear_registry(self):
        for table in self.registry.execute("SELECT name FROM sqlite_schema WHERE type='table';").fetchall():
            if table["name"] == "sqlite_sequence":
                continue
            self.registry.execute(f"DROP TABLE {table['name']};")
            self.registry.connection.commit()

    def install(self):
        # yeah, really.. __builtins__ sometimes appears as a dict and other times as a module, don't ask me why
        if isinstance(__builtins__, dict):
            __builtins__["use"] = self
        elif isinstance(__builtins__, ModuleType):
            setattr(__builtins__, "use", self)
        else:
            raise RuntimeWarning("__builtins__ is something unexpected")

    def uninstall(self):
        if isinstance(__builtins__, dict):
            if "use" in __builtins__:
                del __builtins__["use"]
        elif isinstance(__builtins__, ModuleType):
            if hasattr(__builtins__, "use"):
                delattr(__builtins__, "use")
        else:
            raise RuntimeWarning("__builtins__ is something unexpected")

    def del_entry(self, name, version):
        # TODO: CASCADE to artifacts etc
        self.registry.execute(
            "DELETE FROM hashes WHERE artifact_id IN (SELECT id FROM artifacts WHERE distribution_id IN (SELECT id FROM distributions WHERE name=? AND version=?))",
            (name, str(version)),
        )
        self.registry.execute(
            "DELETE FROM artifacts WHERE distribution_id IN (SELECT id FROM distributions WHERE name=? AND version=?)",
            (name, str(version)),
        )
        self.registry.execute("DELETE FROM distributions WHERE name=? AND version=?", (name, str(version)))
        self.registry.connection.commit()

    def cleanup(self):
        """Bring registry and downloaded packages in sync.

        First all packages are removed that don't have a matching registry entry, then all registry entries that don't have a matching pkg.
        """

        def delete_folder(path):
            for sub in path.iterdir():
                if sub.is_dir():
                    delete_folder(sub)
                else:
                    sub.unlink()
            path.rmdir()

        for name, version, artifact_path, installation_path in self.registry.execute(
            "SELECT name, version, artifact_path, installation_path FROM distributions JOIN artifacts on distributions.id = distribution_id"
        ).fetchall():
            if not (_ensure_path(artifact_path).exists() and _ensure_path(installation_path).exists()):
                self.del_entry(name, version)
        self.registry.connection.commit()

    def _set_mod(self, *, name, mod, path=None, spec=None):
        """Helper to get the order right."""
        self._using[name] = ModInUse(name, mod, path, spec)

    @methdispatch
    def __call__(self, thing, /, *args, **kwargs):
        raise NotImplementedError(UserMessage.cant_use(thing))

    @require(lambda hash_algo: hash_algo in Hash)
    @require(lambda as_import: as_import.isidentifier())
    @__call__.register(URL)
    def _use_url(
        self,
        url: URL,
        /,
        *,
        hash_algo=Hash.sha256,
        hash_value=None,
        initial_globals: Optional[dict[Any, Any]] = None,
        as_import: str = None,
        default=Modes.fastfail,
        modes=0,
    ) -> ProxyModule:
        log.debug(f"use-url: {url}")
        reckless = Modes.recklessness & modes

        response = requests.get(str(url))
        if response.status_code != 200:
            raise ImportError(UserMessage.web_error(url, response))
        this_hash = hash_algo.value(response.content).hexdigest()

        if hash_value and not reckless:
            if this_hash != hash_value:
                return _fail_or_default(
                    UnexpectedHash(f"{this_hash} does not match the expected hash {hash_value} - aborting!"),
                    default,
                )
        else:
            warn(UserMessage.no_validation(url, hash_algo, this_hash), NoValidationWarning)

        name = url.path.segments[-1]
        try:
            mod = _build_mod(
                module_name=name,
                code=response.content,
                module_path=_ensure_path(url.path),
                initial_globals=initial_globals,
            )
        except KeyError:
            raise
        if exc := None:
            return _fail_or_default(ImportError(exc), default)

        frame = inspect.getframeinfo(inspect.currentframe())
        self._set_mod(name=name, mod=mod)
        if as_import:
            sys.modules[as_import] = mod
        return ProxyModule(mod)

    @__call__.register(git)
    def _use_git(
        self,
        git_repo: git,
        /,
        *,
        modes=0,
    ) -> ProxyModule:
        """Install git repo."""

    @__call__.register(Path)
    def _use_path(
        self,
        path,
        /,
        *,
        initial_globals=None,
        as_import: str = None,
        default=Modes.fastfail,
        modes=0,
    ) -> ProxyModule:
        """Import a module from a path.

        https://github.com/amogorkon/justuse/wiki/Use-Path

        Args:
                path ([type]): must be a pathlib.Path
                initial_globals ([type], optional): dict that should be globally available to the module before executing it. Defaults to None.
                default ([type], optional): Return instead if an exception is encountered.
                modes (int, optional): [description]. Defaults to 0; Acceptable mode for this variant: use.reloading.

        Returns:
                Optional[ModuleType]: The module if it was imported, otherwise whatever was specified as default.
        """
        initial_globals = initial_globals or {}
        if as_import:
            assert as_import.isidentifier()

        reloading = bool(Use.reloading & modes)

        exc = None
        mod = None
        original_cwd = Path.cwd()

        if path.is_dir():
            return _fail_or_default(ImportError(f"Can't import directory {path}"), default)

        try:
            name, module_name, package_name, path = _real_path(
                path=path, _applied_decorators=_applied_decorators, landmark=Use.__call__.__code__
            )
        except (NotImplementedError, ImportError):
            exc = traceback.format_exc()

        with open(path, "rb") as rfile:
            code = rfile.read()
        try:
            mod = _build_mod(
                module_name=module_name,
                code=code,
                initial_globals=initial_globals,
                module_path=path,
                package_name=package_name,
            )
        except KeyError:
            exc = traceback.format_exc()
        if exc:
            return _fail_or_default(exc, default)
        mod = ProxyModule(mod)

        if reloading:
            reloader = ModuleReloader(
                proxy=mod,
                name=name,
                path=path,
                initial_globals=initial_globals,
                package_name=package_name,
            )
            _reloaders[mod] = reloader

            threaded = False
            # this looks like a hack, but isn't one -
            # jupyter is running an async loop internally, which works better async than threaded!
            try:
                asyncio.get_running_loop()
            # we're dealing with non-async code, we need threading
            except RuntimeError:
                # can't have the code inside the handler because of "during handling of X, another exception Y happened"
                threaded = True
            if not threaded:
                reloader.start_async()
            else:
                reloader.start_threaded()

            if not all(
                inspect.isfunction(value)
                for key, value in mod.__dict__.items()
                if key not in initial_globals.keys() and not key.startswith("__")
            ):
                warn(UserMessage.not_reloadable(name), NotReloadableWarning)

        os.chdir(original_cwd)
        if exc:
            return _fail_or_default(ImportError(exc), default)

        if as_import:
            sys.modules[as_import] = mod
        self._set_mod(name=name, mod=mod)
        return mod

    @__call__.register(type(None))  # singledispatch is picky - can't be anything but a type
    def _use_kwargs(
        self,
        _: None,  # sic! otherwise single-dispatch with 'empty' *args won't work
        /,
        *,
        package_name: str = None,
        module_name: str = None,
        version: Optional[Union[Version, str]] = None,
        hash_algo=Hash.sha256,
        hashes: Optional[Union[str, list[str]]] = None,
        default=Modes.fastfail,
        modes: int = 0,
    ) -> ProxyModule:
        """
        Import a pkg by name.

        https://github.com/amogorkon/justuse/wiki/Use-String

        Args:
                name (str): The name of the pkg to import.
                version (str or Version, optional): The version of the pkg to import. Defaults to None.
                hash_algo (member of Use.Hash, optional): For future compatibility with more modern hashing algorithms. Defaults to Hash.sha256.
                hashes (str | [str]), optional): A single hash or list of hashes of the pkg to import. Defaults to None.
                default (anything, optional): Whatever should be returned in case there's a problem with the import. Defaults to mode.fastfail.
                modes (int, optional): Any combination of Use.modes . Defaults to 0.

        Raises:
                RuntimeWarning: May be raised if the auto-installation of the pkg fails for some reason.

        Returns:
                ProxyModule|Any: Module if successful, default as specified otherwise.
        """
        log.debug(f"use-kwargs: {package_name} {module_name} {version} {hashes}")
        return self._use_package(
            name=f"{package_name}/{module_name}",
            package_name=package_name,
            module_name=module_name,
            version=Version(version) if version else None,
            hash_algo=hash_algo,
            hashes=hashes,
            default=default,
            modes=modes,
            Message=KwargMessage,
        )

    @__call__.register(tuple)
    def _use_tuple(
        self,
        pkg_tuple,
        /,
        *,
        version: Optional[Union[Version, str]] = None,
        hash_algo=Hash.sha256,
        hashes: Optional[Union[str, list[str]]] = None,
        default=Modes.fastfail,
        modes: int = 0,
    ) -> ProxyModule:
        """
        Import a pkg by name.

        https://github.com/amogorkon/justuse/wiki/Use-String

        Args:
                name (str): The name of the pkg to import.
                version (str or Version, optional): The version of the pkg to import. Defaults to None.
                hash_algo (member of Use.Hash, optional): For future compatibility with more modern hashing algorithms. Defaults to Hash.sha256.
                hashes (str | [str]), optional): A single hash or list of hashes of the pkg to import. Defaults to None.
                default (anything, optional): Whatever should be returned in case there's a problem with the import. Defaults to mode.fastfail.
                modes (int, optional): Any combination of Use.modes . Defaults to 0.

        Raises:
                RuntimeWarning: May be raised if the auto-installation of the pkg fails for some reason.

        Returns:
                ProxyModule|Any: Module if successful, default as specified otherwise.
        """
        log.debug(f"use-tuple: {pkg_tuple} {version} {hashes}")
        package_name, module_name = pkg_tuple
        return self._use_package(
            name=f"{package_name}/{module_name}",
            package_name=package_name,
            module_name=module_name,
            version=Version(version) if version else None,
            hash_algo=hash_algo,
            hashes=hashes,
            default=default,
            modes=modes,
            Message=TupleMessage,
        )

    @__call__.register(str)
    def _use_str(
        self,
        name: str,
        /,
        *,
        version: Optional[Union[Version, str]] = None,
        hash_algo=Hash.sha256,
        hashes: Optional[Union[str, list[str]]] = None,
        default=Modes.fastfail,
        modes: int = 0,
    ) -> ProxyModule:
        """
        Import a pkg by name.

        https://github.com/amogorkon/justuse/wiki/Use-String

        Args:
                name (str): The name of the pkg to import.
                version (str or Version, optional): The version of the pkg to import. Defaults to None.
                hash_algo (member of Use.Hash, optional): For future compatibility with more modern hashing algorithms. Defaults to Hash.sha256.
                hashes (str | [str]), optional): A single hash or list of hashes of the pkg to import. Defaults to None.
                default (anything, optional): Whatever should be returned in case there's a problem with the import. Defaults to Modes.fastfail.
                modes (int, optional): Any combination of Use.modes . Defaults to 0.

        Raises:
                RuntimeWarning: May be raised if something non-critical happens during import.
                ImportError: May be raised if the auto-installation of the pkg fails for some reason.

        Returns:
                ProxyModule|Any: Module (wrapped in a ProxyModule) if successful, default as specified if the requested Module couldn't be imported for some reason.
        """
        package_name, module_name = _parse_name(name)
        return self._use_package(
            name=name,
            package_name=package_name,
            module_name=module_name,
            version=Version(version) if version else None,
            hash_algo=hash_algo,
            hashes=hashes,
            default=default,
            modes=modes,
            Message=StrMessage,
        )

    @require(lambda hash_algo: hash_algo != None)
    def _use_package(
        self,
        *,
        name,
        package_name: str,
        module_name: str,
        version: Optional[Version],
        hashes: Optional[Union[str, set]],
        default: Any,
        hash_algo: Hash,
        modes: int = 0,
        Message: type = UserMessage,
    ):
        auto_install = bool(Modes.auto_install & modes)
        no_public_installation = bool(Modes.no_public_installation & modes)
        fastfail = bool(Modes.fastfail & modes)
        fatal_exceptions = bool(Modes.fatal_exceptions & modes)
        no_browser = bool(Modes.no_browser & modes)
        cleanup = not bool(Modes.no_cleanup & modes)

        if module_name:
            module_name = module_name.replace("/", ".").replace("-", "_")

        # a single hash is a string
        if isinstance(hashes, str):
            # spaces are hard to see in JACK, so we ignore them
            hashes = ["".join(hashes.split())]
        if not hashes:
            hashes = set()
        hashes: set[int] = {
            JACK_as_num(H) if is_JACK(H) else int(H, 16) for H in ("".join(H.split()) for H in hashes)
        }

        # The "try and guess" behaviour is due to how classical imports work,
        # which is inherently ambiguous, but can't really be avoided for packages.
        # let's first see if the user might mean something else entirely
        if _ensure_path(f"./{module_name}.py").exists():
            warn(Message.ambiguous_name_warning(name), AmbiguityWarning)
        spec = None

        if name in self._using:
            spec = self._using[name].spec
        elif not auto_install:
            spec = importlib.util.find_spec(module_name.replace("-", "_"))

        case = bool(version), bool(hashes), bool(spec), auto_install
        log.info("case = %s", case)
        # welcome to the buffet table, where everything is a lie
        kwargs = {
            "name": name,
            "package_name": package_name,
            "module_name": module_name,
            "version": version,
            "user_provided_hashes": hashes,
            "hash_algo": hash_algo,
            "spec": spec,
            "fastfail": fastfail,
            "no_public_installation": no_public_installation,
            "fatal_exceptions": fatal_exceptions,
            "sys_version": Version(".".join(map(str, sys.version_info[:3]))),
            "no_browser": no_browser,
            "Message": Message,
            "registry": self.registry,
            "cleanup": cleanup,
        }

        result = buffet_table(case, kwargs)
        assert result
        assert isinstance(result, (Exception, ModuleType))

        if isinstance(result, Exception):
            return _fail_or_default(result, default)

        if isinstance(result, ModuleType):
            self._set_mod(name=name, mod=result, spec=spec)
            return ProxyModule(result)



"""
Buffet pattern.
Basically dispatch on a specific case, pass in all local context from which the function takes what it needs.
"""

from logging import getLogger

log = getLogger(__name__)

from use import pimp
from use.messages import UserMessage as Message

# the buffet of the past (<3.10)

# fmt: off
def buffet_table(case, kwargs):
    name = kwargs.get('name')
    case_func = {
        (0, 0, 0, 0): lambda: ImportError(Message.cant_import(name)),
        (0, 0, 0, 1): lambda: pimp._pebkac_no_version_no_hash(**kwargs),
        (0, 0, 1, 0): lambda: pimp._import_public_no_install(**kwargs),
        (0, 1, 0, 0): lambda: ImportError(Message.cant_import(name)),
        (1, 0, 0, 0): lambda: ImportError(Message.cant_import(name)),
        (0, 0, 1, 1): lambda: pimp._pebkac_no_version_no_hash(**kwargs),
        (0, 1, 1, 0): lambda: pimp._import_public_no_install(**kwargs),
        (1, 1, 0, 0): lambda: ImportError(Message.cant_import(name)),
        (1, 0, 0, 1): lambda: pimp._pebkac_no_hash(**kwargs),
        (1, 0, 1, 0): lambda: pimp._ensure_version(pimp._import_public_no_install(**kwargs), **kwargs),
        (0, 1, 0, 1): lambda: pimp._pebkac_no_version(**kwargs),
        (0, 1, 1, 1): lambda: pimp._pebkac_no_version(**kwargs),
        (1, 0, 1, 1): lambda: pimp._pebkac_no_hash(**kwargs),
        (1, 1, 0, 1): lambda: pimp._auto_install(**kwargs),
        (1, 1, 1, 0): lambda: pimp._ensure_version(pimp._import_public_no_install(**kwargs), **kwargs),
        (1, 1, 1, 1): lambda: pimp._auto_install(
            func=lambda: pimp._ensure_version(pimp._import_public_no_install(**kwargs), **kwargs), **kwargs
        ),
    }[case]
    result = case_func()
    log.info("result = %s", repr(result))
    return result
# fmt: on
